#!/usr/bin/env python3

# Copyright (C) 2011-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2011-2024, Damon Lynch"

import contextlib
import errno
import hashlib
import locale
import logging
import os
import pickle
import shutil
import stat
import sys
from collections import defaultdict
from itertools import chain
from operator import attrgetter

with contextlib.suppress(locale.Error):
    # Use the default locale as defined by the LANG variable
    locale.setlocale(locale.LC_ALL, "")

import gphoto2 as gp

from raphodo.camera import Camera, CameraProblemEx, gphoto2_python_logging
from raphodo.constants import CameraErrorCode, DownloadStatus, FileType
from raphodo.interprocess import (
    CopyFilesArguments,  # noqa: F401
    CopyFilesResults,
    WorkerInPublishPullPipeline,
)
from raphodo.prefs.preferences import Preferences
from raphodo.problemnotification import (
    CameraFileReadProblem,
    CameraInitializationProblem,
    CopyingProblems,
    FileCopyProblem,
    FileDeleteProblem,
    FileMoveProblem,
    FileWriteProblem,
)
from raphodo.rescan import RescanCamera
from raphodo.rpdfile import RPDFile
from raphodo.storage.storage import get_uri
from raphodo.utilities import GenerateRandomFileName, create_temp_dirs, same_device


def copy_file_metadata(src: str, dst: str) -> tuple | None:
    """
    Copy all stat info (mode bits, atime, mtime, flags) from src to
    dst.

    Adapted from python's shutil.copystat().

    Necessary because with some NTFS file systems, there can be
    problems with setting filesystem metadata like permissions and
    modification time

    :return Tuple of errors, if there were any, else None
    """

    st = os.stat(src)
    mode = stat.S_IMODE(st.st_mode)
    errors = []

    try:
        os.utime(dst, (st.st_atime, st.st_mtime))
    except (OSError, PermissionError, FileNotFoundError) as inst:
        errors.append(inst)

    try:
        os.chmod(dst, mode)
    except (OSError, PermissionError, FileNotFoundError) as inst:
        errors.append(inst)

    if hasattr(os, "chflags") and hasattr(st, "st_flags"):
        try:
            os.chflags(dst, st.st_flags)
        except OSError as why:
            for err in "EOPNOTSUPP", "ENOTSUP":
                if hasattr(errno, err) and why.errno == getattr(errno, err):
                    break
            else:
                pass

    if errors:
        return tuple(errors)

    # Test code:
    # try:
    #     os.chown('/', 1000, 1000)
    # except OSError as inst:
    #     return inst,


def copy_camera_file_metadata(mtime: float, dst: str) -> tuple | None:
    # test code:
    # try:
    #     os.chown('/', 1000, 1000)
    # except OSError as inst:
    #     return inst,

    try:
        os.utime(dst, (mtime, mtime))
    except (OSError, PermissionError, FileNotFoundError) as inst:
        return (inst,)  # note the comma: return a Tuple


class FileCopy:
    """
    Used by classes CopyFilesWorker and BackupFilesWorker
    """

    def __init__(self):
        self.io_buffer = 1024 * 1024
        self.batch_size_bytes = 5 * 1024 * 1024
        self.dest = self.src = None

        self.bytes_downloaded = 0
        self.total_downloaded = 0

    def cleanup_pre_stop(self):
        if self.dest is not None:
            self.dest.close()
        if self.src is not None:
            self.src.close()

    def init_copy_progress(self) -> None:
        self.bytes_downloaded = 0

    def copy_from_filesystem(
        self, source: str, destination: str, rpd_file: RPDFile
    ) -> bool:
        src_chunks = []
        try:
            with open(destination, "wb", self.io_buffer) as self.dest, open(
                source, "rb", self.io_buffer
            ) as self.src:
                total = rpd_file.size
                amount_downloaded = 0

                while True:
                    # first check if process is being stopped or paused
                    self.check_for_controller_directive()

                    chunk = self.src.read(self.io_buffer)
                    if chunk:
                        self.dest.write(chunk)
                        if self.verify_file:
                            src_chunks.append(chunk)
                        amount_downloaded += len(chunk)
                        self.update_progress(amount_downloaded, total)
                    else:
                        break

            if self.verify_file:
                src_bytes = b"".join(src_chunks)
                rpd_file.md5 = hashlib.md5(src_bytes).hexdigest()

            return True
        except (OSError, FileNotFoundError, PermissionError) as e:
            self.problems.append(
                FileCopyProblem(
                    name=os.path.basename(source),
                    uri=get_uri(full_file_name=source),
                    exception=e,
                )
            )
            try:
                msg = f"{e.errno}: {e.strerror}"
            except AttributeError:
                msg = str(e)
            logging.error("%s. Failed to copy %s to %s", msg, source, destination)
            return False
        except Exception as e:
            self.problems.append(
                FileCopyProblem(
                    name=os.path.basename(source),
                    uri=get_uri(full_file_name=source),
                    exception=e,
                )
            )
            try:
                msg = f"{e.errno}: {e.strerror}"
            except AttributeError:
                msg = str(e)
            logging.error(
                "Unexpected error: %s. Failed to copy %s to %s",
                msg,
                source,
                destination,
            )
            return False


class CopyFilesWorker(WorkerInPublishPullPipeline, FileCopy):
    def __init__(self):
        super().__init__("CopyFiles")

    def terminate_camera_removed(self) -> None:
        self.cleanup_pre_stop()
        self.content = pickle.dumps(
            CopyFilesResults(scan_id=self.scan_id, camera_removed=True),
            pickle.HIGHEST_PROTOCOL,
        )
        self.send_message_to_sink()
        self.disconnect_logging()
        self.send_finished_command()
        sys.exit(0)

    def cleanup_pre_stop(self) -> None:
        super().cleanup_pre_stop()
        if self.camera is not None and self.camera.camera_initialized:
            self.camera.free_camera()
        self.send_problems()

    def send_problems(self) -> None:
        """
        Send problems encountered copying to the main process.

        Always sends problems, even if empty, because of the
        possibility that there were filesystem metadata errors
        encountered.
        """

        self.content = pickle.dumps(
            CopyFilesResults(scan_id=self.scan_id, problems=self.problems),
            pickle.HIGHEST_PROTOCOL,
        )
        self.send_message_to_sink()

    def update_progress(self, amount_downloaded: int, total: int) -> None:
        """
        Update the main process about how many bytes have been copied

        :param amount_downloaded: the size in bytes of the file that
         has been copied
        :param total: the size of the file in bytes
        """

        chunk_downloaded = amount_downloaded - self.bytes_downloaded
        if (chunk_downloaded > self.batch_size_bytes) or (amount_downloaded == total):
            self.bytes_downloaded = amount_downloaded
            self.content = pickle.dumps(
                CopyFilesResults(
                    scan_id=self.scan_id,
                    total_downloaded=self.total_downloaded + amount_downloaded,
                    chunk_downloaded=chunk_downloaded,
                ),
                pickle.HIGHEST_PROTOCOL,
            )
            self.send_message_to_sink()

            # if amount_downloaded == total:
            #     self.bytes_downloaded = 0

    def copy_from_camera(self, rpd_file: RPDFile) -> bool:
        try:
            src_bytes = self.camera.save_file_by_chunks(
                dir_name=rpd_file.path,
                file_name=rpd_file.name,
                size=rpd_file.size,
                dest_full_filename=rpd_file.temp_full_file_name,
                progress_callback=self.update_progress,
                check_for_command=self.check_for_controller_directive,
                return_file_bytes=self.verify_file,
            )
        except CameraProblemEx as e:
            name = rpd_file.name
            uri = rpd_file.get_uri()
            if e.gp_code in (gp.GP_ERROR_IO_USB_FIND, gp.GP_ERROR_BAD_PARAMETERS):
                self.terminate_camera_removed()
            elif e.code == CameraErrorCode.read:
                self.problems.append(
                    CameraFileReadProblem(name=name, uri=uri, gp_code=e.gp_code)
                )
            else:
                assert e.code == CameraErrorCode.write
                self.problems.append(
                    FileWriteProblem(name=name, uri=uri, exception=e.py_exception)
                )
            return False

        if self.verify_file:
            rpd_file.md5 = hashlib.md5(src_bytes).hexdigest()

        return True

    def copy_associate_file(
        self,
        rpd_file: RPDFile,
        temp_name: str,
        dest_dir: str,
        associate_file_fullname: str,
        file_type: str,
    ) -> str | None:
        ext = os.path.splitext(associate_file_fullname)[1]
        temp_ext = f"{temp_name}{ext}"
        temp_full_name = os.path.join(dest_dir, temp_ext)
        if rpd_file.from_camera:
            dir_name, file_name = os.path.split(associate_file_fullname)
            try:
                self.camera.save_file(dir_name, file_name, temp_full_name)
            except CameraProblemEx as e:
                uri = get_uri(
                    full_file_name=associate_file_fullname,
                    camera_details=rpd_file.camera_details,
                )
                if e.gp_code in (gp.GP_ERROR_IO_USB_FIND, gp.GP_ERROR_BAD_PARAMETERS):
                    self.terminate_camera_removed()
                elif e.code == CameraErrorCode.read:
                    self.problems.append(
                        CameraFileReadProblem(
                            name=file_name, uri=uri, gp_code=e.gp_code
                        )
                    )
                else:
                    assert e.code == CameraErrorCode.write
                    self.problems.append(
                        FileWriteProblem(
                            name=file_name, uri=uri, exception=e.py_exception
                        )
                    )
                logging.error(
                    "Failed to download %s file: %s", file_type, associate_file_fullname
                )
                return None
        else:
            try:
                shutil.copyfile(associate_file_fullname, temp_full_name)
            except (OSError, FileNotFoundError, PermissionError) as e:
                logging.error(
                    "Failed to download %s file: %s", file_type, associate_file_fullname
                )
                logging.error("%s: %s", e.errno, e.strerror)
                name = os.path.basename(associate_file_fullname)
                uri = get_uri(full_file_name=associate_file_fullname)
                self.problems.append(FileWriteProblem(name=name, uri=uri, exception=e))
                return None
            logging.debug("Copied %s file %s", file_type, temp_full_name)

        # Adjust file modification times and other file system metadata
        # Ignore any errors copying file system metadata -- assume they would
        # have been raised when copying the primary file's filesystem metadata
        if rpd_file.from_camera:
            copy_camera_file_metadata(
                mtime=rpd_file.modification_time, dst=temp_full_name
            )
        else:
            copy_file_metadata(associate_file_fullname, temp_full_name)
        return temp_full_name

    def do_work(self):
        self.problems = CopyingProblems()
        args = pickle.loads(self.content)  # type: CopyFilesArguments

        if args.log_gphoto2:
            self.gphoto2_logging = gphoto2_python_logging()

        self.scan_id = args.scan_id
        self.verify_file = args.verify_file

        self.camera = None

        # To workaround a bug in iOS and possibly other devices, check if need to
        # rescan the files on the device
        rescan_check = [
            rpd_file
            for rpd_file in args.files
            if rpd_file.from_camera and not rpd_file.cache_full_file_name
        ]
        no_rescan = [
            rpd_file
            for rpd_file in args.files
            if not rpd_file.from_camera or rpd_file.cache_full_file_name
        ]

        if rescan_check:
            prefs = Preferences()
            # Initialize camera
            try:
                self.camera = Camera(
                    model=args.device.camera_model,
                    port=args.device.camera_port,
                    is_mtp_device=args.device.is_mtp_device,
                    raise_errors=True,
                    specific_folders=prefs.folders_to_scan,
                )
            except CameraProblemEx as e:
                self.problems.append(CameraInitializationProblem(gp_code=e.gp_code))
                logging.error(
                    "Could not initialize camera %s %s",
                    args.device.camera_model,
                    args.device.camera_port,
                )
                self.terminate_camera_removed()
            else:
                rescan = RescanCamera(camera=self.camera, prefs=prefs)
                rescan.rescan_camera(rpd_files=rescan_check)
                rescan_check = rescan.rpd_files
                if rescan.missing_rpd_files:
                    logging.error(
                        "%s files could not be relocated on %s",
                        len(rescan.missing_rpd_files),
                        self.camera.display_name,
                    )
                    rescan_check = list(chain(rescan_check, rescan.missing_rpd_files))

        rpd_files = list(chain(rescan_check, no_rescan))

        random_filename = GenerateRandomFileName()

        rpd_cache_same_device = defaultdict(lambda: None)  # type: defaultdict[FileType, bool|None]

        photo_temp_dir, video_temp_dir = create_temp_dirs(
            args.photo_download_folder, args.video_download_folder
        )

        # Notify main process of temp directory names
        self.content = pickle.dumps(
            CopyFilesResults(
                scan_id=args.scan_id,
                photo_temp_dir=photo_temp_dir or "",
                video_temp_dir=video_temp_dir or "",
            ),
            pickle.HIGHEST_PROTOCOL,
        )
        self.send_message_to_sink()

        # Sort the files to be copied by modification time
        # Important to do this with respect to sequence numbers, or else
        # they'll be downloaded in what looks like a random order
        rpd_files = sorted(rpd_files, key=attrgetter("modification_time"))

        self.display_name = args.device.display_name

        for idx, rpd_file in enumerate(rpd_files):
            self.dest = self.src = None

            if rpd_file.file_type == FileType.photo:
                dest_dir = photo_temp_dir
            else:
                dest_dir = video_temp_dir

            # Three scenarios:
            # 1. Downloading from device with file system we can directly
            #    access
            # 2. Downloading from camera using libgphoto2
            # 3. Downloading from camera where we've already cached at
            #    least some of the files in the Download Cache

            self.init_copy_progress()

            if rpd_file.cache_full_file_name and os.path.isfile(
                rpd_file.cache_full_file_name
            ):
                # Scenario 3
                temp_file_name = os.path.basename(rpd_file.cache_full_file_name)
                temp_name = os.path.splitext(temp_file_name)[0]
                temp_full_file_name = os.path.join(dest_dir, temp_file_name)

                if rpd_cache_same_device[rpd_file.file_type] is None:
                    rpd_cache_same_device[rpd_file.file_type] = same_device(
                        rpd_file.cache_full_file_name, dest_dir
                    )

                if rpd_cache_same_device[rpd_file.file_type]:
                    try:
                        shutil.move(rpd_file.cache_full_file_name, temp_full_file_name)
                        copy_succeeded = True
                    except (OSError, PermissionError, FileNotFoundError) as inst:
                        copy_succeeded = False
                        logging.error(
                            "Could not move cached file %s to temporary file %s. Error "
                            "code: %s",
                            rpd_file.cache_full_file_name,
                            temp_full_file_name,
                            inst.errno,
                        )
                        self.problems.append(
                            FileMoveProblem(
                                name=rpd_file.name,
                                uri=rpd_file.get_uri(),
                                exception=inst,
                            )
                        )
                    if self.verify_file:
                        with open(temp_full_file_name, "rb") as f:
                            rpd_file.md5 = hashlib.md5(f.read()).hexdigest()
                    self.update_progress(rpd_file.size, rpd_file.size)
                else:
                    # The download folder changed since the scan occurred, and is now
                    # on a different file system compared to that where the devices
                    # files were cached. Or the file was downloaded in full by the scan
                    # stage and saved, e.g. a sample video.
                    source = rpd_file.cache_full_file_name
                    destination = temp_full_file_name
                    copy_succeeded = self.copy_from_filesystem(
                        source, destination, rpd_file
                    )
                    try:
                        os.remove(source)
                    except (OSError, PermissionError, FileNotFoundError) as e:
                        logging.error(
                            "Error removing RPD Cache file %s while copying %s. "
                            "Error code: %s",
                            source,
                            rpd_file.full_file_name,
                            e.errno,
                        )
                        self.problems.append(
                            FileDeleteProblem(
                                name=os.path.basename(source),
                                uri=get_uri(source),
                                exception=e,
                            )
                        )

            else:
                # Scenario 1 or 2
                # Generate temporary name 5 digits long, because we cannot
                # guarantee the source does not have duplicate file names in
                # different directories, and here we are copying the files into
                # a single directory
                temp_name = random_filename.name()
                temp_name_ext = f"{temp_name}.{rpd_file.extension}"
                temp_full_file_name = os.path.join(dest_dir, temp_name_ext)

            rpd_file.temp_full_file_name = temp_full_file_name

            if not rpd_file.cache_full_file_name:
                if rpd_file.from_camera:
                    # Scenario 2
                    if not self.camera:
                        copy_succeeded = False
                        logging.error(
                            "Could not copy %s from the %s",
                            rpd_file.full_file_name,
                            self.display_name,
                        )
                        self.update_progress(rpd_file.size, rpd_file.size)
                    else:
                        copy_succeeded = self.copy_from_camera(rpd_file)
                else:
                    # Scenario 1
                    source = rpd_file.full_file_name
                    destination = rpd_file.temp_full_file_name
                    copy_succeeded = self.copy_from_filesystem(
                        source, destination, rpd_file
                    )

            # increment this amount regardless of whether the copy actually
            # succeeded or not. It's necessary to keep the user informed.
            self.total_downloaded += rpd_file.size

            mdata_exceptions = None

            if not copy_succeeded:
                rpd_file.status = DownloadStatus.download_failed
                logging.debug("Download failed for %s", rpd_file.full_file_name)
            else:
                if rpd_file.from_camera:
                    mdata_exceptions = copy_camera_file_metadata(
                        float(rpd_file.modification_time), temp_full_file_name
                    )
                else:
                    mdata_exceptions = copy_file_metadata(
                        rpd_file.full_file_name, temp_full_file_name
                    )

                # copy THM (video thumbnail file) if there is one
                if rpd_file.thm_full_name:
                    rpd_file.temp_thm_full_name = self.copy_associate_file(
                        # translators: refers to the video thumbnail file that some
                        # cameras generate -- it has a .THM file extension
                        rpd_file,
                        temp_name,
                        dest_dir,
                        rpd_file.thm_full_name,
                        _("video THM"),
                    )

                # copy audio file if there is one
                if rpd_file.audio_file_full_name:
                    rpd_file.temp_audio_full_name = self.copy_associate_file(
                        rpd_file,
                        temp_name,
                        dest_dir,
                        rpd_file.audio_file_full_name,
                        _("audio"),
                    )

                # copy XMP file if there is one
                if rpd_file.xmp_file_full_name:
                    rpd_file.temp_xmp_full_name = self.copy_associate_file(
                        rpd_file,
                        temp_name,
                        dest_dir,
                        rpd_file.xmp_file_full_name,
                        "XMP",
                    )

                # copy Magic Lantern LOG file if there is one
                if rpd_file.log_file_full_name:
                    rpd_file.temp_log_full_name = self.copy_associate_file(
                        rpd_file,
                        temp_name,
                        dest_dir,
                        rpd_file.log_file_full_name,
                        "LOG",
                    )

            download_count = idx + 1

            self.content = pickle.dumps(
                CopyFilesResults(
                    copy_succeeded=copy_succeeded,
                    rpd_file=rpd_file,
                    download_count=download_count,
                    mdata_exceptions=mdata_exceptions,
                ),
                pickle.HIGHEST_PROTOCOL,
            )
            self.send_message_to_sink()

        if len(self.problems):
            logging.debug(
                "Encountered %s problems while copying from %s",
                len(self.problems),
                self.display_name,
            )
        self.send_problems()

        if self.camera is not None:
            self.camera.free_camera()

        self.disconnect_logging()
        self.send_finished_command()


if __name__ == "__main__":
    copy = CopyFilesWorker()



#!/usr/bin/env python3

# Copyright (C) 2011-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Generates names for files and folders, and renames (moves) files.

Runs as a daemon process.
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2011-2024, Damon Lynch"

import contextlib
import errno
import locale
import logging
import os
import pickle
import sqlite3
import sys
from collections import namedtuple
from datetime import datetime
from enum import Enum

with contextlib.suppress(locale.Error):
    # Use the default locale as defined by the LANG variable
    locale.setlocale(locale.LC_ALL, "")


import raphodo.generatename as gn
import raphodo.metadata.exiftool as exiftool
from raphodo.constants import (
    ConflictResolution,
    DownloadStatus,
    FileType,
    RenameAndMoveStatus,
)
from raphodo.interprocess import (
    DaemonProcess,
    RenameAndMoveFileData,  # noqa: F401
    RenameAndMoveFileResults,
)
from raphodo.prefs.preferences import DownloadsTodayTracker, Preferences
from raphodo.problemnotification import (
    DuplicateFileWhenSyncingProblem,
    FileAlreadyExistsProblem,
    FileMetadataLoadProblem,
    IdentifierAddedProblem,
    NoDataToNameProblem,
    RenamingAssociateFileProblem,
    RenamingFileProblem,
    RenamingProblems,
    SameNameDifferentExif,
    SubfolderCreationProblem,
    make_href,
)
from raphodo.rpdfile import Photo, RPDFile, Video
from raphodo.rpdsql import DownloadedSQL
from raphodo.storage.storage import get_uri
from raphodo.utilities import (
    datetime_roughly_equal,
    platform_c_maxint,
    stdchannel_redirected,
)


class SyncRawJpegStatus(Enum):
    matching_pair = 1
    no_match = 2
    error_already_downloaded = 3
    error_datetime_mismatch = 4


SyncRawJpegMatch = namedtuple("SyncRawJpegMatch", "status, sequence_number")
SyncRawJpegResult = namedtuple(
    "SyncRawJpegResult", "sequence_to_use, failed, photo_name, " "photo_ext"
)
SyncRawJpegRecord = namedtuple(
    "SyncRawJpegRecord", "extension, date_time, sequence_number_used"
)


class SyncRawJpeg:
    """
    Match JPEG and RAW images so that they have the same file names
    """

    def __init__(self):
        self.photos = {}  # type: dict[str, SyncRawJpegRecord]

    def add_download(
        self,
        name: str,
        extension: str,
        date_time: datetime,
        sequence_number_used: gn.MatchedSequences,
    ) -> None:
        if not isinstance(date_time, datetime):
            logging.debug(
                "Rejecting %s for sync RAW jpeg matching because its metadata "
                "date time does not exist",
                name,
            )
            return

        if name not in self.photos:
            self.photos[name] = SyncRawJpegRecord(
                extension=[extension],
                date_time=date_time,
                sequence_number_used=sequence_number_used,
            )
        else:
            if extension not in self.photos[name].extension:
                self.photos[name].extension.append(extension)

    def matching_pair(
        self, name: str, extension: str, date_time: datetime
    ) -> SyncRawJpegMatch:
        """
        Checks to see if the image matches an image that has already been
        downloaded.

        Image name (minus extension), exif date time are checked. Exif
        date times should be within 30 seconds of each other, because
        need to allow for the fact that RAW and jpegs might not be
        written to the memory card(s) at the same time.

        :return: Returns SyncRawJpegStatus.error_already_downloaded
         and a sequence number if the name, extension, and exif values
         match (i.e. it has already been downloaded).

         Returns SyncRawJpegStatus.matching_pair and a sequence number
         if name and exif values match, but the extension is different
         (i.e. a matching RAW + JPG image).

         Returns SyncRawJpegStatus.error_datetime_mismatch and a
         sequence number of None if photos detected with the same
         filenames, but taken at different times.

         Returns SyncRawJpegStatus.no_match and a sequence number
         of None if no match
        """

        if name in self.photos:
            if datetime_roughly_equal(self.photos[name].date_time, date_time, 30):
                if extension in self.photos[name].extension:
                    return SyncRawJpegMatch(
                        SyncRawJpegStatus.error_already_downloaded,
                        self.photos[name].sequence_number_used,
                    )
                else:
                    return SyncRawJpegMatch(
                        SyncRawJpegStatus.matching_pair,
                        self.photos[name].sequence_number_used,
                    )
            else:
                return SyncRawJpegMatch(SyncRawJpegStatus.error_datetime_mismatch, None)
        return SyncRawJpegMatch(SyncRawJpegStatus.no_match, None)

    def ext_exif_date_time(self, name) -> tuple[str, datetime]:
        """
        Returns first extension, and exif date time data for
        the already downloaded photo
        """

        return self.photos[name].extension[0], self.photos[name].date_time


def load_metadata(
    rpd_file: Photo | Video,
    et_process: exiftool.ExifTool,
    problems: RenamingProblems,
    force_exiftool: bool,
) -> bool:
    """
    Loads the metadata for the file.

    :param rpd_file: photo or video
    :param et_process: the daemon ExifTool process
    :param problems: problems encountered renaming the file
    :param force_exiftool: whether to force the use of ExifTool to read the file
     metadata
    :return True if operation succeeded, false otherwise
    """
    if rpd_file.metadata is None and not rpd_file.load_metadata(
        full_file_name=rpd_file.temp_full_file_name,
        et_process=et_process,
        force_exiftool=force_exiftool,
    ):
        # Error in reading metadata

        problems.append(
            FileMetadataLoadProblem(
                name=rpd_file.name, uri=rpd_file.get_uri(), file_type=rpd_file.title
            )
        )
        return False
    return True


def _generate_name(
    generator: gn.PhotoName | gn.PhotoSubfolder | gn.VideoName | gn.VideoSubfolder,
    rpd_file: Photo | Video,
    et_process: exiftool.ExifTool,
    problems: RenamingProblems,
    force_exiftool: bool,
) -> str:
    """
    Generate a subfolder or file name.

    :param generator: subfolder or file name generator, appropriate
     for the file type (photo or video)
    :param rpd_file: file to work on
    :param et_process:  the daemon ExifTool process
    :param problems: problems encountered renaming the file
    :param force_exiftool: whether to force the use of ExifTool to read the file
     metadata
    :return: the name in string format, emptry string if error
    """
    do_generation = load_metadata(
        rpd_file=rpd_file,
        et_process=et_process,
        problems=problems,
        force_exiftool=force_exiftool,
    )

    if do_generation:
        value = generator.generate_name(rpd_file)
        if value is None:
            value = ""
    else:
        value = ""

    return value


def generate_subfolder(
    rpd_file: Photo | Video,
    et_process: exiftool.ExifTool,
    problems: RenamingProblems,
    force_exiftool: bool,
) -> None:
    """
    Generate subfolder names e.g. 2015/201512

    :param rpd_file: file to work on
    :param et_process:  the daemon ExifTool process
    :param problems: problems encountered renaming the file
    :param force_exiftool: whether to force the use of ExifTool to read the file
     metadata
    """

    if rpd_file.file_type == FileType.photo:
        generator = gn.PhotoSubfolder(rpd_file.subfolder_pref_list, problems=problems)
    else:
        generator = gn.VideoSubfolder(rpd_file.subfolder_pref_list, problems=problems)

    rpd_file.download_subfolder = _generate_name(
        generator=generator,
        rpd_file=rpd_file,
        et_process=et_process,
        problems=problems,
        force_exiftool=force_exiftool,
    )


def generate_name(
    rpd_file: Photo | Video,
    et_process: exiftool.ExifTool,
    problems: RenamingProblems,
    force_exiftool: bool,
) -> None:
    """
    Generate file names e.g. 20150607-1.cr2

    :param rpd_file: file to work on
    :param et_process:  the daemon ExifTool process
    :param problems: problems encountered renaming the file
    :param force_exiftool: whether to force the use of ExifTool to read the file
     metadata
    """

    if rpd_file.file_type == FileType.photo:
        generator = gn.PhotoName(pref_list=rpd_file.name_pref_list, problems=problems)
    else:
        generator = gn.VideoName(pref_list=rpd_file.name_pref_list, problems=problems)

    rpd_file.download_name = _generate_name(
        generator=generator,
        rpd_file=rpd_file,
        et_process=et_process,
        problems=problems,
        force_exiftool=force_exiftool,
    )


class RenameMoveFileWorker(DaemonProcess):
    """
    Generates names for files and folders, and renames (moves) files.

    Runs as a daemon process.
    """

    def __init__(self) -> None:
        super().__init__("Rename and Move")

        self.prefs = Preferences()

        self.sync_raw_jpeg = SyncRawJpeg()
        self.downloaded = DownloadedSQL()

        logging.debug("Start of day is set to %s", self.prefs.day_start)

        self.platform_c_maxint = platform_c_maxint()

        # This will be assigned again in run(), but initializing it here
        # clarifies any problems with type checking in an IDE
        self.problems = RenamingProblems()

    def notify_file_already_exists(
        self, rpd_file: Photo | Video, identifier: str | None = None
    ) -> None:
        """
        Notify user that the download file already exists
        """

        # get information on when the existing file was last modified
        try:
            modification_time = os.path.getmtime(rpd_file.download_full_file_name)
            dt = datetime.fromtimestamp(modification_time)
            date = dt.strftime("%x")
            time = dt.strftime("%X")
        except Exception:
            logging.error(
                "Could not determine the file modification time of %s",
                rpd_file.download_full_file_name,
            )
            date = time = ""

        source = rpd_file.get_souce_href()

        device = make_href(name=rpd_file.device_display_name, uri=rpd_file.device_uri)

        if not identifier:
            problem = FileAlreadyExistsProblem(
                file_type_capitalized=rpd_file.title_capitalized,
                file_type=rpd_file.title,
                name=rpd_file.download_name,
                uri=get_uri(full_file_name=rpd_file.download_full_file_name),
                source=source,
                device=device,
                date=date,
                time=time,
            )

            rpd_file.status = DownloadStatus.download_failed
        else:
            problem = IdentifierAddedProblem(
                file_type_capitalized=rpd_file.title_capitalized,
                file_type=rpd_file.title,
                name=rpd_file.download_name,
                uri=get_uri(full_file_name=rpd_file.download_full_file_name),
                source=source,
                device=device,
                date=date,
                time=time,
                identifier=identifier,
            )

            rpd_file.status = DownloadStatus.downloaded_with_warning

        self.problems.append(problem)

    def notify_download_failure_file_error(
        self, rpd_file: Photo | Video, inst: Exception
    ) -> None:
        """
        Handle cases where file failed to download
        """
        uri = get_uri(
            full_file_name=rpd_file.full_file_name,
            camera_details=rpd_file.camera_details,
        )
        device = make_href(name=rpd_file.device_display_name, uri=rpd_file.device_uri)

        problem = RenamingFileProblem(
            file_type=rpd_file.title,
            destination=rpd_file.download_name,
            folder=rpd_file.download_path,
            name=rpd_file.name,
            uri=uri,
            device=device,
            exception=inst,
        )
        self.problems.append(problem)

        rpd_file.status = DownloadStatus.download_failed

        try:
            msg = "Failed to create file {}: {} {}".format(
                rpd_file.download_full_file_name, inst.errno, inst.strerror
            )
            logging.error(msg)
        except AttributeError:
            logging.error(
                "Failed to create file %s: %s ",
                rpd_file.download_full_file_name,
                str(inst),
            )

    def download_file_exists(self, rpd_file: Photo | Video) -> bool:
        """
        Check how to handle a download file already existing
        """

        if self.prefs.conflict_resolution == ConflictResolution.add_identifier:
            logging.debug(
                "Will add unique identifier to avoid duplicate filename for %s",
                rpd_file.full_file_name,
            )
            return True
        else:
            self.notify_file_already_exists(rpd_file)
            return False

    def same_name_different_exif(
        self, sync_photo_name: str, rpd_file: Photo | Video
    ) -> None:
        """
        Notify the user that a file was already downloaded with the same
        name, but the exif information was different
        """

        i1_ext, i1_date_time = self.sync_raw_jpeg.ext_exif_date_time(sync_photo_name)
        image2_date_time = rpd_file.date_time()
        assert isinstance(i1_date_time, datetime)
        i1_date = i1_date_time.strftime("%x")
        i1_time = i1_date_time.strftime("%X")
        assert isinstance(image2_date_time, datetime)
        image2_date = image2_date_time.strftime("%x")
        image2_time = image2_date_time.strftime("%X")

        self.problems.append(
            SameNameDifferentExif(
                image1=f"{sync_photo_name}{i1_ext}",
                image1_date=i1_date,
                image1_time=i1_time,
                image2=rpd_file.name,
                image2_date=image2_date,
                image2_time=image2_time,
            )
        )
        rpd_file.status = DownloadStatus.downloaded_with_warning

    def _move_associate_file(
        self, extension: str, full_base_name: str, temp_associate_file: str
    ) -> str:
        """
        Move (rename) the associate file using the pre-generated name.

        Exceptions are not caught.

        :return: full path and filename
        """

        download_full_name = full_base_name + extension

        # move (rename) associate file
        # don't check to see if it already exists
        os.rename(temp_associate_file, download_full_name)

        return download_full_name

    def move_thm_file(self, rpd_file: Photo | Video) -> None:
        """
        Move (rename) the THM thumbnail file using the pre-generated name
        """

        try:
            ext = rpd_file.thm_extension if rpd_file.thm_extension else ".THM"
        except AttributeError:
            ext = ".THM"

        try:
            rpd_file.download_thm_full_name = self._move_associate_file(
                extension=ext,
                full_base_name=rpd_file.download_full_base_name,
                temp_associate_file=rpd_file.temp_thm_full_name,
            )
        except (OSError, FileNotFoundError) as e:
            self.problems.append(
                RenamingAssociateFileProblem(
                    source=make_href(
                        name=os.path.basename(rpd_file.download_thm_full_name),
                        uri=get_uri(
                            full_file_name=rpd_file.download_thm_full_name,
                            camera_details=rpd_file.camera_details,
                        ),
                    ),
                    exception=e,
                )
            )

            logging.error(
                "Failed to move video THM file %s", rpd_file.download_thm_full_name
            )

    def move_audio_file(self, rpd_file: Photo | Video) -> None:
        """
        Move (rename) the associate audio file using the pre-generated
        name
        """

        try:
            ext = rpd_file.audio_extension if rpd_file.audio_extension else ".WAV"
        except AttributeError:
            ext = ".WAV"

        try:
            rpd_file.download_audio_full_name = self._move_associate_file(
                extension=ext,
                full_base_name=rpd_file.download_full_base_name,
                temp_associate_file=rpd_file.temp_audio_full_name,
            )
        except (OSError, FileNotFoundError) as e:
            self.problems.append(
                RenamingAssociateFileProblem(
                    source=make_href(
                        name=os.path.basename(rpd_file.download_audio_full_name),
                        uri=get_uri(
                            full_file_name=rpd_file.download_audio_full_name,
                            camera_details=rpd_file.camera_details,
                        ),
                    ),
                    exception=e,
                )
            )
            logging.error(
                "Failed to move file's associated audio file %s",
                rpd_file.download_audio_full_name,
            )

    def move_xmp_file(self, rpd_file: Photo | Video) -> None:
        """
        Move (rename) the associate XMP file using the pre-generated
        name
        """

        try:
            ext = rpd_file.xmp_extension if rpd_file.xmp_extension else ".XMP"
        except AttributeError:
            ext = ".XMP"

        try:
            rpd_file.download_xmp_full_name = self._move_associate_file(
                extension=ext,
                full_base_name=rpd_file.download_full_base_name,
                temp_associate_file=rpd_file.temp_xmp_full_name,
            )
        except (OSError, FileNotFoundError) as e:
            self.problems.append(
                RenamingAssociateFileProblem(
                    source=make_href(
                        name=os.path.basename(rpd_file.download_xmp_full_name),
                        uri=get_uri(
                            full_file_name=rpd_file.download_xmp_full_name,
                            camera_details=rpd_file.camera_details,
                        ),
                    ),
                    exception=e,
                )
            )
            logging.error(
                "Failed to move file's associated XMP file %s",
                rpd_file.download_xmp_full_name,
            )

    def move_log_file(self, rpd_file: Photo | Video) -> None:
        """
        Move (rename) the associate XMP file using the pre-generated
        name
        """

        try:
            ext = rpd_file.log_extension if rpd_file.log_extension else ".LOG"
        except AttributeError:
            ext = ".LOG"

        try:
            rpd_file.download_log_full_name = self._move_associate_file(
                extension=ext,
                full_base_name=rpd_file.download_full_base_name,
                temp_associate_file=rpd_file.temp_log_full_name,
            )
        except (OSError, FileNotFoundError) as e:
            self.problems.append(
                RenamingAssociateFileProblem(
                    source=make_href(
                        name=os.path.basename(rpd_file.download_log_full_name),
                        uri=get_uri(
                            full_file_name=rpd_file.download_log_full_name,
                            camera_details=rpd_file.camera_details,
                        ),
                    ),
                    exception=e,
                )
            )
            logging.error(
                "Failed to move file's associated LOG file %s",
                rpd_file.download_log_full_name,
            )

    def check_for_fatal_name_generation_errors(self, rpd_file: Photo | Video) -> bool:
        """
        :return False if either the download subfolder or filename are
         blank, else returns True
        """

        if not rpd_file.download_subfolder or not rpd_file.download_name:
            if not rpd_file.download_subfolder and not rpd_file.download_name:
                area = _("subfolder and filename")
            elif not rpd_file.download_name:
                area = _("filename")
            else:
                area = _("subfolder")

            rpd_file.status = DownloadStatus.download_failed
            self.problems.append(
                NoDataToNameProblem(
                    name=rpd_file.name,
                    uri=rpd_file.get_uri(),
                    area=area,
                    file_type=rpd_file.title,
                )
            )
            return False
        else:
            return True

    def add_unique_identifier(self, rpd_file: Photo | Video) -> bool:
        """
        Adds a unique identifier like _1 to a filename, in ever
        incrementing values, until a unique filename is generated.

        :param rpd_file: the file being worked on
        :return: True if the operation was successful, else returns
         False
        """

        name = os.path.splitext(rpd_file.download_name)
        full_name = rpd_file.download_full_file_name
        while True:
            self.duplicate_files[full_name] = self.duplicate_files.get(full_name, 0) + 1
            identifier = "_%s" % self.duplicate_files[full_name]
            rpd_file.download_name = f"{name[0]}{identifier}{name[1]}"
            rpd_file.download_full_file_name = os.path.join(
                rpd_file.download_path, rpd_file.download_name
            )

            try:
                if os.path.exists(rpd_file.download_full_file_name):
                    raise OSError(
                        errno.EEXIST,
                        "File exists: %s" % rpd_file.download_full_file_name,
                    )
                os.rename(
                    rpd_file.temp_full_file_name, rpd_file.download_full_file_name
                )
                self.notify_file_already_exists(rpd_file, identifier)
                return True

            except OSError as inst:
                if inst.errno != errno.EEXIST:
                    self.notify_download_failure_file_error(rpd_file, inst)
                    return False

    def sync_raw_jpg(self, rpd_file: Photo | Video) -> SyncRawJpegResult:
        failed = False
        sequence_to_use = None
        photo_name, photo_ext = os.path.splitext(rpd_file.name)
        if not load_metadata(
            rpd_file=rpd_file,
            et_process=self.exiftool_process,
            problems=self.problems,
            force_exiftool=self.prefs.force_exiftool,
        ):
            failed = True
            rpd_file.status = DownloadStatus.download_failed
            self.check_for_fatal_name_generation_errors(rpd_file)
        else:
            date_time = rpd_file.date_time()
            if not isinstance(date_time, datetime):
                failed = True
                rpd_file.status = DownloadStatus.download_failed
                self.check_for_fatal_name_generation_errors(rpd_file)
            else:
                matching_pair = self.sync_raw_jpeg.matching_pair(
                    name=photo_name, extension=photo_ext, date_time=date_time
                )  # type: SyncRawJpegMatch
                sequence_to_use = matching_pair.sequence_number
                if matching_pair.status == SyncRawJpegStatus.error_already_downloaded:
                    # this exact file has already been
                    # downloaded (same extension, same filename,
                    # and roughly the same exif date time  info)
                    if (
                        self.prefs.conflict_resolution
                        != ConflictResolution.add_identifier
                    ):
                        self.problems.append(
                            DuplicateFileWhenSyncingProblem(
                                name=rpd_file.name,
                                uri=rpd_file.get_uri(),
                                file_type=rpd_file.title,
                            )
                        )

                        rpd_file.status = DownloadStatus.download_failed
                        failed = True
                else:
                    self.sequences.matched_sequences = matching_pair.sequence_number
                    self.sequences.use_matched_sequences = (
                        self.sequences.matched_sequences is not None
                    )
                    if (
                        matching_pair.status
                        == SyncRawJpegStatus.error_datetime_mismatch
                    ):
                        self.same_name_different_exif(photo_name, rpd_file)

        return SyncRawJpegResult(sequence_to_use, failed, photo_name, photo_ext)

    def prepare_rpd_file(self, rpd_file: Photo | Video) -> None:
        """
        Populate the RPDFile with download values used in subfolder
        and filename generation
        """

        if rpd_file.file_type == FileType.photo:
            rpd_file.download_folder = self.prefs.photo_download_folder
            rpd_file.subfolder_pref_list = self.prefs.photo_subfolder
            rpd_file.name_pref_list = self.prefs.photo_rename
        else:
            rpd_file.download_folder = self.prefs.video_download_folder
            rpd_file.subfolder_pref_list = self.prefs.video_subfolder
            rpd_file.name_pref_list = self.prefs.video_rename

    def process_rename_failure(self, rpd_file: RPDFile) -> None:
        try:
            os.remove(rpd_file.temp_full_file_name)
        except OSError:
            logging.error(
                "Failed to delete temporary file %s", rpd_file.temp_full_file_name
            )

    def generate_names(
        self, rpd_file: Photo | Video, synchronize_raw_jpg: bool
    ) -> bool:
        rpd_file.strip_characters = self.prefs.strip_characters

        generate_subfolder(
            rpd_file=rpd_file,
            et_process=self.exiftool_process,
            problems=self.problems,
            force_exiftool=self.prefs.force_exiftool,
        )

        if rpd_file.download_subfolder:
            logging.debug(
                "Generated subfolder name %s for file %s",
                rpd_file.download_subfolder,
                rpd_file.name,
            )

            self.sequences.stored_sequence_no = self.prefs.stored_sequence_no
            if self.must_synchronize_raw_jpg and rpd_file.file_type == FileType.video:
                self.sequences.use_matched_sequences = False

            rpd_file.sequences = self.sequences

            # generate the file name
            generate_name(
                rpd_file=rpd_file,
                et_process=self.exiftool_process,
                problems=self.problems,
                force_exiftool=self.prefs.force_exiftool,
            )

            if rpd_file.name_generation_problem:
                logging.warning(
                    "Encountered a problem generating file name for file %s",
                    rpd_file.name,
                )
                rpd_file.status = DownloadStatus.downloaded_with_warning
            else:
                logging.debug(
                    "Generated file name %s for file %s",
                    rpd_file.download_name,
                    rpd_file.name,
                )
        else:
            logging.error(
                "Failed to generate subfolder name for file: %s", rpd_file.name
            )

        return self.check_for_fatal_name_generation_errors(rpd_file)

    def move_file(self, rpd_file: Photo | Video) -> bool:
        """
        Having generated the file name and subfolder names, move
        the file
        :param rpd_file: photo or video being worked on
        :return: True if move succeeded, False otherwise
        """

        move_succeeded = False

        rpd_file.download_path = os.path.join(
            rpd_file.download_folder, rpd_file.download_subfolder
        )
        rpd_file.download_full_file_name = os.path.join(
            rpd_file.download_path, rpd_file.download_name
        )
        rpd_file.download_full_base_name = os.path.splitext(
            rpd_file.download_full_file_name
        )[0]

        if not os.path.isdir(rpd_file.download_path):
            try:
                os.makedirs(rpd_file.download_path)
            except OSError as inst:
                if inst.errno != errno.EEXIST:
                    logging.error(
                        "Failed to create download subfolder: %s",
                        rpd_file.download_path,
                    )
                    logging.error(inst)

                    problem = SubfolderCreationProblem(
                        folder=make_href(
                            name=rpd_file.download_subfolder,
                            uri=get_uri(path=rpd_file.download_path),
                        ),
                        exception=inst,
                    )
                    self.problems.append(problem)

        # Move temp file to subfolder

        add_unique_identifier = False
        try:
            if os.path.exists(rpd_file.download_full_file_name):
                raise OSError(
                    errno.EEXIST, "File exists: %s" % rpd_file.download_full_file_name
                )
            logging.debug(
                "Renaming %s to %s .....",
                rpd_file.temp_full_file_name,
                rpd_file.download_full_file_name,
            )
            os.rename(rpd_file.temp_full_file_name, rpd_file.download_full_file_name)
            logging.debug("....successfully renamed file")
            move_succeeded = True
            if rpd_file.status != DownloadStatus.downloaded_with_warning:
                rpd_file.status = DownloadStatus.downloaded
        except OSError as inst:
            if inst.errno == errno.EEXIST:
                add_unique_identifier = self.download_file_exists(rpd_file)
            else:
                self.notify_download_failure_file_error(rpd_file, inst)
        except Exception as inst:
            # all other errors, including PermissionError
            self.notify_download_failure_file_error(rpd_file, inst)

        if add_unique_identifier:
            move_succeeded = self.add_unique_identifier(rpd_file)

        return move_succeeded

    def process_file(self, rpd_file: Photo | Video, download_count: int) -> bool:
        """
        Generate file & subfolder name, and move (rename) photo / video
        :param rpd_file: photo or video
        :param download_count: used to track the file being downloaded via a counter
        :return: success or otherwise of operation
        """

        move_succeeded = False

        self.prepare_rpd_file(rpd_file)

        synchronize_raw_jpg = (
            self.must_synchronize_raw_jpg and rpd_file.file_type == FileType.photo
        )

        if synchronize_raw_jpg:
            sync_result = self.sync_raw_jpg(rpd_file)

            if sync_result.failed:
                return False

        generation_succeeded = self.generate_names(rpd_file, synchronize_raw_jpg)

        if generation_succeeded:
            move_succeeded = self.move_file(rpd_file)

            logging.debug("Finished processing file: %s", download_count)

        if move_succeeded:
            if synchronize_raw_jpg:
                if sync_result.sequence_to_use is None:
                    sequence = self.sequences.create_matched_sequences()
                else:
                    sequence = sync_result.sequence_to_use
                self.sync_raw_jpeg.add_download(
                    name=sync_result.photo_name,
                    extension=sync_result.photo_ext,
                    date_time=rpd_file.date_time(),
                    sequence_number_used=sequence,
                )

            if not synchronize_raw_jpg or (
                synchronize_raw_jpg and sync_result.sequence_to_use is None
            ):
                if self.uses_sequence_session_no or self.uses_sequence_letter:
                    self.sequences.increment(
                        self.uses_sequence_session_no, self.uses_sequence_letter
                    )
                if self.uses_stored_sequence_no:
                    if self.prefs.stored_sequence_no == self.platform_c_maxint:
                        # wrap value if it exceeds the maximum size value that Qt can
                        # display in its spinbox
                        self.prefs.stored_sequence_no = 0
                    else:
                        self.prefs.stored_sequence_no += 1
                self.downloads_today_tracker.increment_downloads_today()

            if rpd_file.temp_thm_full_name:
                self.move_thm_file(rpd_file)

            if rpd_file.temp_audio_full_name:
                self.move_audio_file(rpd_file)

            if rpd_file.temp_xmp_full_name:
                self.move_xmp_file(rpd_file)

            if rpd_file.temp_log_full_name:
                self.move_log_file(rpd_file)

        return move_succeeded

    def initialise_downloads_today_stored_number(self) -> None:
        """
        Initialize (or reinitialize) Downloads Today and Stored No
        sequence values from the program preferences.
        """

        # Synchronize QSettings instance in preferences class
        self.prefs.sync()
        if self.prefs.any_pref_uses_stored_sequence_no():
            logging.info(
                "Stored number at download start: %s", self.prefs.stored_sequence_no + 1
            )

        # Track downloads today, using a class whose purpose is to
        # take the value in the user prefs, increment, and then
        # finally used to update the prefs
        self.downloads_today_tracker = DownloadsTodayTracker(
            day_start=self.prefs.day_start, downloads_today=self.prefs.downloads_today
        )

    def initialise_sequence_number_usage(self) -> None:
        """
        Determine what type of sequence numbers are being used in file name generation
        """

        self.uses_sequence_session_no = self.prefs.any_pref_uses_session_sequence_no()
        self.uses_sequence_letter = self.prefs.any_pref_uses_sequence_letter_value()
        self.uses_stored_sequence_no = self.prefs.any_pref_uses_stored_sequence_no()

    def run(self) -> None:
        """
        Generate subfolder and filename, and attempt to move the file
        from its temporary directory.

        Move video THM and/or audio file if there is one.

        If successful, increment sequence values.

        Report any success or failure.
        """
        i = 0

        # Dict of filename keys and int values used to track ints to add as
        # suffixes to duplicate files
        self.duplicate_files = {}

        self.initialise_downloads_today_stored_number()

        self.sequences = gn.Sequences(
            self.downloads_today_tracker, self.prefs.stored_sequence_no
        )

        with stdchannel_redirected(
            sys.stderr, os.devnull
        ), exiftool.ExifTool() as self.exiftool_process:
            while True:
                if i:
                    logging.debug("Finished %s. Getting next task.", i)

                # rename file and move to generated subfolder
                directive, content = self.receiver.recv_multipart()

                self.check_for_command(directive, content)

                data = pickle.loads(content)  # type: RenameAndMoveFileData
                if data.message == RenameAndMoveStatus.download_started:
                    # reinitialize downloads today and stored sequence number
                    # in case the user has updated them via the user interface
                    self.initialise_downloads_today_stored_number()
                    self.sequences.downloads_today_tracker = (
                        self.downloads_today_tracker
                    )
                    self.sequences.stored_sequence_no = self.prefs.stored_sequence_no

                    dl_today = (
                        self.downloads_today_tracker.get_or_reset_downloads_today()
                    )
                    logging.debug("Completed downloads today: %s", dl_today)

                    self.initialise_sequence_number_usage()

                    self.must_synchronize_raw_jpg = (
                        self.prefs.must_synchronize_raw_jpg()
                    )

                    self.problems = RenamingProblems()

                elif data.message == RenameAndMoveStatus.download_completed:
                    if len(self.problems):
                        self.content = pickle.dumps(
                            RenameAndMoveFileResults(problems=self.problems),
                            pickle.HIGHEST_PROTOCOL,
                        )
                        self.send_message_to_sink()

                    # Ask main application process to update prefs with stored
                    # sequence number and downloads today values. But first sync
                    # the prefs here, to write out the dirty values so they are not
                    # saved when a sync is done at download start, overwriting
                    # the values that may have been changed in the main process
                    logging.debug(
                        "Rename and move process syncing preferences to the file "
                        "system"
                    )
                    self.prefs.sync()
                    self.content = pickle.dumps(
                        RenameAndMoveFileResults(
                            stored_sequence_no=self.sequences.stored_sequence_no,
                            downloads_today=self.downloads_today_tracker.downloads_today,
                        ),
                        pickle.HIGHEST_PROTOCOL,
                    )
                    dl_today = (
                        self.downloads_today_tracker.get_or_reset_downloads_today()
                    )
                    logging.debug("Downloads today: %s", dl_today)
                    self.send_message_to_sink()
                else:
                    rpd_file = data.rpd_file
                    download_count = data.download_count

                    if data.download_succeeded:
                        move_succeeded = self.process_file(rpd_file, download_count)
                        if not move_succeeded:
                            self.process_rename_failure(rpd_file)
                        else:
                            # Record file as downloaded in SQLite database
                            try:
                                self.downloaded.add_downloaded_file(
                                    name=rpd_file.name,
                                    size=rpd_file.size,
                                    modification_time=rpd_file.modification_time,
                                    download_full_file_name=rpd_file.download_full_file_name,
                                )
                            except sqlite3.OperationalError as e:
                                # This should never happen because this is the only
                                # process writing to the database..... but just in
                                # case
                                logging.error(
                                    "Database error adding download file %s: %s. "
                                    "Will not retry.",
                                    rpd_file.download_full_file_name,
                                    e,
                                )
                    else:
                        move_succeeded = False

                    rpd_file.metadata = None
                    self.content = pickle.dumps(
                        RenameAndMoveFileResults(
                            move_succeeded=move_succeeded,
                            rpd_file=rpd_file,
                            download_count=download_count,
                        ),
                        pickle.HIGHEST_PROTOCOL,
                    )
                    self.send_message_to_sink()

                    i += 1


if __name__ == "__main__":
    rename = RenameMoveFileWorker()
    rename.run()



# Copyright (C) 2016-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader. If not,
# see <http://www.gnu.org/licenses/>.


__author__ = "Damon Lynch"
__copyright__ = "Copyright 2016-2024, Damon Lynch"


import io
import logging
import os
import traceback

from PyQt5.QtWidgets import QApplication, QMessageBox

try:
    from easygui import codebox

    have_easygui = True
except Exception:
    # if import failed for any reason, ignore it
    have_easygui = False

from raphodo.iplogging import full_log_file_path
from raphodo.prefs.preferences import Preferences
from raphodo.storage.storage import get_uri
from raphodo.ui.viewutils import standardMessageBox
from raphodo.utilities import bug_report_full_tar_path, create_bugreport_tar

message_box_displayed = False
exceptions_notified = set()

# Translators: do not translate the HTML tags such as  <a> or <br>, or the Python
# string formatting tags such as website.
please_report_problem_body = _(
    'Please report the problem at <a href="{website}">{website}</a>.<br><br>'
    "In your bug report describe what you expected to happen, and what you observed "
    "happening.<br><br>"
    "The bug report must also include the program settings and log files. To create a "
    "file with this additional information, click Save."
)

tar_created_title = _("Additional Information Saved")

# Translators: do not translate the HTML tags such as <pre>, <a>, or <br>, or the Python
# string formatting tags tarfile and uri.
tar_created_body = _(
    "The additional bug report information was created in your home directory in "
    "a tar file: <pre>{tarfile}</pre>"
    "You need to attach this file to the bug report yourself. It will not be "
    "automatically attached.<br><br>"
    'Click <a href="{uri}">here</a> to see the file in your file manager.'
)

tar_error_title = _("Error Creating Additional Information")

tar_error_header = _(
    "The additional bug report information was not created. Please file a bug report "
    "anyway."
)

# Translators: do not translate the HTML tags such as <i>, <a>, or <br>, or the Python
# string formatting tags log_file, etc.
tar_error_body = _(
    "Include in your bug report the program's log files. The bug report must include "
    "<i>{log_file}</i>, but attaching the other log files is often helpful.<br><br>"
    "If possible, please also include the program's configuration file "
    "<i>{config_file}</i>.<br><br>"
    'Click <a href="{log_path}">here</a> to open the log directory, and '
    '<a href="{config_path}">here</a> to open the configuration directory.'
)

upgrade_message = _(
    'Upgrading to the <a href="{website}">latest version</a> will allow you to '
    "determine if the problem you encountered has already been fixed."
)


def save_bug_report_tar(config_file: str, full_log_file_path: str) -> None:
    """
    Save a tar file in the user's home directory with logging files and config file.
    Inform the user of the result using QMessageBox.

    :param config_file: full path to the config file
    :param full_log_file_path: full path to the directory with the log files
    """

    bug_report_full_tar = bug_report_full_tar_path()

    logging.info("Creating bug report tar file %s", bug_report_full_tar)
    log_path, log_file = os.path.split(full_log_file_path)
    if create_bugreport_tar(
        full_tar_name=bug_report_full_tar,
        log_path=log_path,
        full_config_file=config_file,
    ):

        body = tar_created_body.format(
            tarfile=os.path.split(bug_report_full_tar)[1],
            uri=get_uri(full_file_name=bug_report_full_tar),
        )
        messagebox = standardMessageBox(
            message=body,
            rich_text=True,
            title=tar_created_title,
            standardButtons=QMessageBox.Ok,
        )
        messagebox.exec_()
    else:
        # There was some kind of problem generating the tar file, e.g. no free space
        log_uri = get_uri(log_path)
        config_path, config_file = os.path.split(config_file)
        config_uri = get_uri(path=config_path)

        body = tar_error_body.format(
            log_path=log_uri,
            log_file=log_file,
            config_path=config_uri,
            config_file=config_file,
        )
        message = f"<b>{tar_error_header}</b><br><br>{body}"
        messageBox = standardMessageBox(
            message=message,
            rich_text=True,
            title=tar_error_title,
            standardButtons=QMessageBox.Ok,
        )
        messageBox.exec_()


def excepthook(exception_type, exception_value, traceback_object) -> None:
    """
    Global function to catch unhandled exceptions.

    Inspired by function of the same name in the Eric project, but subsequently heavily
    modified.
    """

    if traceback_object is not None:
        frame = traceback_object.tb_frame
        filename = frame.f_code.co_filename
        lineno = traceback_object.tb_lineno
    else:
        lineno = -1
        filename = "unknown"
    key = f"{filename}{lineno}"

    global message_box_displayed

    tb_file = io.StringIO()
    traceback.print_exception(
        exception_type, exception_value, traceback_object, limit=None, file=tb_file
    )
    tb_file.seek(0)
    traceback_info = tb_file.read()

    logging.error("An unhandled exception occurred")
    logging.error(traceback_info)

    if not message_box_displayed and key not in exceptions_notified:
        message_box_displayed = True
        exceptions_notified.add(key)

        prefs = Preferences()

        title = _("Problem in Rapid Photo Downloader")

        if QApplication.instance():

            header = _("A problem occurred in Rapid Photo Downloader")

            only_notification = _(
                "If the same problem occurs again before the program exits, this is "
                "the only notification about it."
            )

            body = please_report_problem_body.format(
                website="https://bugs.rapidphotodownloader.com"
            )

            message = f"<b>{header}</b><br><br>{body}<br><br>{only_notification}"

            errorbox = standardMessageBox(
                message=message,
                rich_text=True,
                title=title,
                standardButtons=QMessageBox.Save | QMessageBox.Cancel,
                defaultButton=QMessageBox.Save,
            )
            errorbox.setDetailedText(traceback_info)
            if errorbox.exec_() == QMessageBox.Save:
                save_bug_report_tar(
                    config_file=prefs.settings_path(),
                    full_log_file_path=full_log_file_path(),
                )

        elif have_easygui:
            message = _("A problem occurred in Rapid Photo Downloader\n")
            prefix = _(
                "Please report the problem at {website}\n"
                "Attach the log file to your bug report, found at {log_path}\n\n"
            ).format(
                website="https://bugs.rapidphotodownloader.com",
                log_path=full_log_file_path(),
            )
            text = prefix + traceback_info
            codebox(msg=message, title=title, text=text)
        message_box_displayed = False



#!/usr/bin/env python3

# Copyright (C) 2007-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.
### USA

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2007-2024, Damon Lynch"

import contextlib
import locale
import logging
import os
import re
from collections import namedtuple
from datetime import datetime, timedelta

with contextlib.suppress(locale.Error):
    # Use the default locale as defined by the LANG variable
    locale.setlocale(locale.LC_ALL, "")


from raphodo.generatenameconfig import (
    APERTURE,
    ARTIST,
    CAMERA_MAKE,
    CAMERA_MODEL,
    CODEC,
    COPYRIGHT,
    DATE_TIME,
    DATE_TIME_CONVERT,
    DOWNLOAD_SEQ_NUMBER,
    DOWNLOAD_TIME,
    EXPOSURE_TIME,
    EXTENSION,
    FILE_NUMBER,
    FILE_NUMBER_FOLDER,
    FILENAME,
    FOCAL_LENGTH,
    FPS,
    HEIGHT,
    IMAGE_DATE,
    IMAGE_NUMBER,
    IMAGE_NUMBER_1,
    IMAGE_NUMBER_2,
    IMAGE_NUMBER_3,
    IMAGE_NUMBER_4,
    IMAGE_NUMBER_ALL,
    ISO,
    JOB_CODE,
    LENGTH,
    LIST_DATE_TIME_L2,
    LIST_SEQUENCE_NUMBERS_L2,
    LIST_SHUTTER_COUNT_L2,
    LOWERCASE,
    METADATA,
    NAME,
    OWNER_NAME,
    SEPARATOR,
    SEQUENCE_LETTER,
    SEQUENCES,
    SERIAL_NUMBER,
    SESSION_SEQ_NUMBER,
    SHORT_CAMERA_MODEL,
    SHORT_CAMERA_MODEL_HYPHEN,
    SHUTTER_COUNT,
    STORED_SEQ_NUMBER,
    SUBSECONDS,
    TEXT,
    TODAY,
    UPPERCASE,
    VIDEO_DATE,
    VIDEO_NUMBER,
    WIDTH,
    YESTERDAY,
    PrefValueInvalidError,
)
from raphodo.prefs.preferences import DownloadsTodayTracker
from raphodo.problemnotification import (
    FilenameNotFullyGeneratedProblem,
    FolderNotFullyGeneratedProblemProblem,
    RenamingProblems,
    make_href,
)
from raphodo.rpdfile import RPDFile
from raphodo.storage.storage import get_uri
from raphodo.utilities import letters

MatchedSequences = namedtuple(
    "MatchedSequences",
    "session_sequence_no, sequence_letter, downloads_today, stored_sequence_no",
)


def convert_date_for_strftime(datetime_user_choice):
    try:
        return DATE_TIME_CONVERT[LIST_DATE_TIME_L2.index(datetime_user_choice)]
    except KeyError:
        raise PrefValueInvalidError(datetime_user_choice)


class abstract_attribute:
    """
    http://stackoverflow.com/questions/32536176/how-to-define-lazy-variable-in-python-which-will-raise-notimplementederror-for-a/32536493
    """

    def __get__(self, obj, type):
        # Now we will iterate over the names on the class,
        # and all its superclasses, and try to find the attribute
        # name for this descriptor
        # traverse the parents in the method resolution order
        for cls in type.__mro__:
            # for each cls thus, see what attributes they set
            for name, value in cls.__dict__.items():
                # we found ourselves here
                if value is self:
                    # if the property gets accessed as Child.variable,
                    # obj will be done. For this case
                    # If accessed as a_child.variable, the class Child is
                    # in the type, and a_child in the obj.
                    this_obj = obj if obj else type

                    raise NotImplementedError(
                        f"{this_obj!r} does not have the attribute {name!r} "
                        f"(abstract from class {cls.__name__!r})"
                    )

        # we did not find a match, should be rare, but prepare for it
        raise NotImplementedError(
            f"{type.__name__} does not set the abstract attribute <unknown>"
        )


GenerationErrors = (
    FilenameNotFullyGeneratedProblem | FolderNotFullyGeneratedProblemProblem
)


class NameGeneration:
    """
    Generate the name of a photo. Used as a base class for generating names
    of videos, as well as subfolder names for both file types
    """

    def __init__(
        self, pref_list: list[str], problems: RenamingProblems | None = None
    ) -> None:
        self.pref_list = pref_list
        self.no_metadata = False

        self.problems = problems
        self.problem = abstract_attribute()  # type: GenerationErrors

        self.strip_forward_slash = abstract_attribute()
        self.add_extension = abstract_attribute()
        self.L1_date_check = abstract_attribute()

        self.L0 = ""
        self.L1 = ""
        self.L2 = ""

    def _get_values_from_pref_list(self):
        for i in range(0, len(self.pref_list), 3):
            yield (self.pref_list[i], self.pref_list[i + 1], self.pref_list[i + 2])

    def _get_date_component(self) -> str:
        """
        Returns portion of new file / subfolder name based on date time.
        If the date is missing, will attempt to use the fallback date.
        """

        # step 1: get the correct value from metadata
        if self.L1 == self.L1_date_check:  # noqa: SIM300
            if self.no_metadata:
                if self.L2 == SUBSECONDS:
                    d = datetime.fromtimestamp(self.rpd_file.modification_time)
                    if not d.microsecond:
                        d = "00"
                    try:
                        d = str(round(int(str(d.microsecond)[:3]) / 10))
                    except Exception:
                        d = "00"
                    return d
                d = datetime.fromtimestamp(self.rpd_file.ctime)
            else:
                if self.L2 == SUBSECONDS:
                    d = self.rpd_file.metadata.sub_seconds(missing=None)
                    if d is None:
                        self.problem.missing_metadata.append(_(self.L2))
                        return ""
                    else:
                        return d
                else:
                    d = self.rpd_file.date_time(missing=None)

        elif self.L1 == TODAY:
            d = datetime.now()
        elif self.L1 == YESTERDAY:
            delta = timedelta(days=1)
            d = datetime.now() - delta
        elif self.L1 == DOWNLOAD_TIME:
            d = self.rpd_file.download_start_time
        else:
            raise TypeError("Date options invalid")

        # step 2: if have a value, try to convert it to string format
        if d:
            try:
                return d.strftime(convert_date_for_strftime(self.L2))
            except Exception as e:
                logging.warning(
                    "Problem converting date/time value for file %s",
                    self.rpd_file.full_file_name,
                )
                self.problem.bad_converstion_date_time = True
                self.problem.bad_conversion_exception = e

        # step 3: handle a missing value using file modification time
        if self.rpd_file.modification_time:
            try:
                d = datetime.fromtimestamp(self.rpd_file.modification_time)
            except Exception:
                logging.error(
                    "Both file modification time and metadata date & time are invalid "
                    "for file %s",
                    self.rpd_file.full_file_name,
                )
                self.problem.invalid_date_time = True
                return ""
        else:
            self.problem.missing_metadata.append(_(self.L1))
            return ""

        try:
            return d.strftime(convert_date_for_strftime(self.L2))
        except Exception:
            logging.error(
                "Both file modification time and metadata date & time are invalid for "
                "file %s",
                self.rpd_file.full_file_name,
            )
            self.problem.invalid_date_time = True
            return ""

    def _get_associated_file_extension(self, associate_file):
        """
        Generates extensions with correct capitalization for files like
        thumbnail or audio files.
        """

        if not associate_file:
            return None

        extension = os.path.splitext(associate_file)[1]
        if self.rpd_file.generate_extension_case == UPPERCASE:
            extension = extension.upper()
        elif self.rpd_file.generate_extension_case == LOWERCASE:
            extension = extension.lower()
        # else keep extension case the same as what it originally was
        return extension

    def _get_thm_extension(self) -> None:
        """
        Generates THM extension with correct capitalization, if needed
        """
        self.rpd_file.thm_extension = self._get_associated_file_extension(
            self.rpd_file.thm_full_name
        )

    def _get_audio_extension(self) -> None:
        """
        Generates audio extension with correct capitalization, if needed
        e.g. WAV or wav
        """
        self.rpd_file.audio_extension = self._get_associated_file_extension(
            self.rpd_file.audio_file_full_name
        )

    def _get_xmp_extension(self) -> None:
        """
        Generates XMP extension with correct capitalization, if needed.
        """

        self.rpd_file.xmp_extension = self._get_associated_file_extension(
            self.rpd_file.xmp_file_full_name
        )

    def _get_log_extension(self) -> None:
        """
        Generates LOG extension with correct capitalization, if needed.
        """

        self.rpd_file.log_extension = self._get_associated_file_extension(
            self.rpd_file.log_file_full_name
        )

    def _get_filename_component(self):
        """
        Returns portion of new file / subfolder name based on the file name
        """

        name, extension = os.path.splitext(self.rpd_file.name)

        if self.L1 == NAME:
            filename = name
        elif self.L1 == EXTENSION:
            # Used in subfolder name generation
            if extension:
                # having the period when this is used as a part of a
                # subfolder name
                # is a bad idea when it is at the start!
                filename = extension[1:]
            else:
                self.problem.missing_extension = True
                return ""
        elif self.L1 == IMAGE_NUMBER or self.L1 == VIDEO_NUMBER:
            n = re.search("(?P<image_number>[0-9]+$)", name)
            if not n:
                self.problem.missing_image_no = True
                return ""
            else:
                image_number = n.group("image_number")

                if self.L2 == IMAGE_NUMBER_ALL:
                    filename = image_number
                elif self.L2 == IMAGE_NUMBER_1:
                    filename = image_number[-1]
                elif self.L2 == IMAGE_NUMBER_2:
                    filename = image_number[-2:]
                elif self.L2 == IMAGE_NUMBER_3:
                    filename = image_number[-3:]
                else:
                    assert self.L2 == IMAGE_NUMBER_4
                    filename = image_number[-4:]
        else:
            raise TypeError("Incorrect filename option")

        if self.L2 == UPPERCASE:
            filename = filename.upper()
        elif self.L2 == LOWERCASE:
            filename = filename.lower()

        return filename

    def _get_metadata_component(self):
        """
        Returns portion of new image / subfolder name based on the metadata

        Note: date time metadata found in _getDateComponent()
        """

        if self.L1 == APERTURE:
            v = self.rpd_file.metadata.aperture()
        elif self.L1 == ISO:
            v = self.rpd_file.metadata.iso()
        elif self.L1 == EXPOSURE_TIME:
            v = self.rpd_file.metadata.exposure_time(alternative_format=True)
        elif self.L1 == FOCAL_LENGTH:
            v = self.rpd_file.metadata.focal_length()
        elif self.L1 == CAMERA_MAKE:
            v = self.rpd_file.metadata.camera_make()
        elif self.L1 == CAMERA_MODEL:
            v = self.rpd_file.metadata.camera_model()
        elif self.L1 == SHORT_CAMERA_MODEL:
            v = self.rpd_file.metadata.short_camera_model()
        elif self.L1 == SHORT_CAMERA_MODEL_HYPHEN:
            v = self.rpd_file.metadata.short_camera_model(include_characters="\-")
        elif self.L1 == SERIAL_NUMBER:
            v = self.rpd_file.metadata.camera_serial()
        elif self.L1 == SHUTTER_COUNT:
            v = self.rpd_file.metadata.shutter_count()
            if v:
                v = int(v)
                padding = LIST_SHUTTER_COUNT_L2.index(self.L2) + 3
                formatter = "%0" + str(padding) + "i"
                v = formatter % v
        elif self.L1 == FILE_NUMBER:
            v = self.rpd_file.metadata.file_number()
            if v and self.L2 == FILE_NUMBER_FOLDER:
                v = v[:3]
        elif self.L1 == OWNER_NAME:
            v = self.rpd_file.metadata.owner_name()
        elif self.L1 == ARTIST:
            v = self.rpd_file.metadata.artist()
        elif self.L1 == COPYRIGHT:
            v = self.rpd_file.metadata.copyright()
        else:
            raise TypeError("Invalid metadata option specified")
        if self.L1 in (
            CAMERA_MAKE,
            CAMERA_MODEL,
            SHORT_CAMERA_MODEL,
            SHORT_CAMERA_MODEL_HYPHEN,
            OWNER_NAME,
            ARTIST,
            COPYRIGHT,
        ):
            if self.L2 == UPPERCASE:
                v = v.upper()
            elif self.L2 == LOWERCASE:
                v = v.lower()
        if not v:
            self.problem.missing_metadata.append(_(self.L1))
        return v

    def _calculate_letter_sequence(self, sequence):
        v = letters(sequence)
        if self.L2 == UPPERCASE:
            v = v.upper()

        return v

    def _format_sequence_no(self, value, amountToPad):
        padding = LIST_SEQUENCE_NUMBERS_L2.index(amountToPad) + 1
        formatter = "%0" + str(padding) + "i"
        return formatter % value

    def _get_downloads_today(self):
        return self._format_sequence_no(
            self.rpd_file.sequences.downloads_today, self.L2
        )

    def _get_session_sequence_no(self):
        return self._format_sequence_no(
            self.rpd_file.sequences.session_sequence_no, self.L2
        )

    def _get_stored_sequence_no(self):
        return self._format_sequence_no(
            self.rpd_file.sequences.stored_sequence_no, self.L2
        )

    def _get_sequence_letter(self):
        return self._calculate_letter_sequence(self.rpd_file.sequences.sequence_letter)

    def _get_sequences_component(self):
        if self.L1 == DOWNLOAD_SEQ_NUMBER:
            return self._get_downloads_today()
        elif self.L1 == SESSION_SEQ_NUMBER:
            return self._get_session_sequence_no()
        elif self.L1 == STORED_SEQ_NUMBER:
            return self._get_stored_sequence_no()
        elif self.L1 == SEQUENCE_LETTER:
            return self._get_sequence_letter()

    def _get_component(self):
        try:
            if self.L0 == DATE_TIME:
                return self._get_date_component()
            elif self.L0 == TEXT:
                return self.L1
            elif self.L0 == FILENAME:
                return self._get_filename_component()
            elif self.L0 == METADATA:
                return self._get_metadata_component()
            elif self.L0 == SEQUENCES:
                return self._get_sequences_component()
            elif self.L0 == JOB_CODE:
                return self.rpd_file.job_code
            elif self.L0 == SEPARATOR:
                return os.sep
        except Exception as e:
            self.problem.component_problem = _(self.L0)
            self.problem.component_exception = e
            return ""

    def filter_strip_characters(self, name: str) -> str:
        """
        Filter out unwanted chacters from file and subfolder names
        :param name: full name or name component
        :return: filtered name
        """

        # remove any null characters - they are bad news in file names
        name = name.replace("\x00", "")

        # the user could potentially copy and paste a block of text with a carriage /
        # line return
        name = name.replace("\n", "")

        if self.rpd_file.strip_characters:
            for c in r'\:*?"<>|':
                name = name.replace(c, "")

        if self.strip_forward_slash:
            name = name.replace("/", "")
        return name

    def _destination(self, rpd_file: RPDFile, name: str) -> str:
        # implement in subclass
        return ""

    def _filter_name(self, name: str, parts: bool) -> str:
        # implement in subclass if need be
        return name

    def generate_name(
        self, rpd_file: RPDFile, parts: bool | None = False
    ) -> str | list[str]:
        """
        Generate subfolder name(s), and photo/video filenames

        :param rpd_file: rpd file for the name to generate
        :param parts: if True, return string components in a list
        :return: complete string or list of name components
        """

        self.rpd_file = rpd_file

        name = [] if parts else ""

        for self.L0, self.L1, self.L2 in self._get_values_from_pref_list():
            v = self._get_component()
            if parts:
                name.append(self.filter_strip_characters(v))
            elif v:
                name += v

        if not parts:
            name = self.filter_strip_characters(name)
            # strip any white space from the beginning and end of the name
            name = name.strip()
        elif name:
            # likewise, strip any white space from the beginning and end of the name
            name[0] = name[0].lstrip()
            name[-1] = name[-1].rstrip()

        if self.add_extension:
            case = rpd_file.generate_extension_case
            extension = os.path.splitext(rpd_file.name)[1]
            if case == UPPERCASE:
                extension = extension.upper()
            elif case == LOWERCASE:
                extension = extension.lower()
            if parts:
                name.append(extension)
            else:
                name += extension

            self._get_thm_extension()
            self._get_audio_extension()
            self._get_xmp_extension()
            self._get_log_extension()

        name = self._filter_name(name, parts)

        if self.problem.has_error():
            rpd_file.name_generation_problem = True

            if self.problems is not None:
                self.problem.destination = self._destination(
                    rpd_file=rpd_file, name=name
                )
                self.problem.file_type = rpd_file.title
                self.problem.source = rpd_file.get_souce_href()
                self.problems.append(self.problem)

        return name


class PhotoName(NameGeneration):
    """
    Generate filenames for photos
    """

    def __init__(
        self, pref_list: list[str], problems: RenamingProblems | None = None
    ) -> None:
        super().__init__(pref_list, problems)

        self.problem = FilenameNotFullyGeneratedProblem()

        self.strip_forward_slash = True
        self.add_extension = True
        self.L1_date_check = IMAGE_DATE  # used in _get_date_component()

    def _destination(self, rpd_file: RPDFile, name: str) -> str:
        if rpd_file.download_subfolder:
            return make_href(
                name=name,
                uri=get_uri(
                    full_file_name=os.path.join(
                        rpd_file.download_folder, rpd_file.download_subfolder, name
                    )
                ),
            )
        else:
            return name


class VideoName(PhotoName):
    """
    Generate filenames for videos
    """

    def __init__(
        self, pref_list: list[str], problems: RenamingProblems | None = None
    ) -> None:
        super().__init__(pref_list, problems)

        self.L1_date_check = VIDEO_DATE  # used in _get_date_component()

    def _get_metadata_component(self):
        """
        Returns portion of video / subfolder name based on the metadata

        Note: date time metadata found in _getDateComponent()
        """
        return get_video_metadata_component(self)


class PhotoSubfolder(NameGeneration):
    """
    Generate subfolder names for photo files
    """

    def __init__(
        self,
        pref_list: list[str],
        problems: RenamingProblems | None = None,
        no_metadata: bool | None = False,
    ) -> None:
        """
        :param pref_list: subfolder generation preferences list
        :param no_metadata: if True, halt as soon as the need for metadata
        or a job code or sequence number becomes necessary
        """

        super().__init__(pref_list, problems)

        if no_metadata:
            self.pref_list = truncate_before_unwanted_subfolder_component(pref_list)
        else:
            self.pref_list = pref_list

        self.no_metadata = no_metadata

        self.problem = FolderNotFullyGeneratedProblemProblem()

        self.strip_extraneous_white_space = re.compile(r"\s*%s\s*" % os.sep)
        self.strip_forward_slash = False
        self.add_extension = False
        self.L1_date_check = IMAGE_DATE  # used in _get_date_component()

    def _filter_name(self, name: str, parts: bool) -> str:
        if not parts:
            return self.filter_subfolder_characters(name)
        return name

    def _destination(self, rpd_file: RPDFile, name: str) -> str:
        return make_href(
            name=name, uri=get_uri(path=os.path.join(rpd_file.download_folder, name))
        )

    def filter_subfolder_characters(self, subfolders: str) -> str:
        """
        Remove unwanted characters specific to the generation of subfolders
        :param subfolders: the complete string containing the subfolders
         (not component parts)
        :return: filtered string
        """

        # subfolder value must never start with a separator, or else any
        # os.path.join function call will fail to join a subfolder to its
        # parent folder
        if subfolders and subfolders[0] == os.sep:
            subfolders = subfolders[1:]

        # remove any spaces before and after a directory name
        if subfolders and self.rpd_file.strip_characters:
            subfolders = self.strip_extraneous_white_space.sub(os.sep, subfolders)

        # remove any repeated directory separators
        double_sep = os.sep * 2
        subfolders = subfolders.replace(double_sep, os.sep)

        # remove any trailing directory separators
        while subfolders.endswith(os.sep):
            subfolders = subfolders[:-1]

        return subfolders


class VideoSubfolder(PhotoSubfolder):
    """
    Generate subfolder names for video files
    """

    def __init__(
        self,
        pref_list: list[str],
        problems: RenamingProblems | None = None,
        no_metadata: bool = False,
    ) -> None:
        """
        :param pref_list: subfolder generation preferences list
        :param no_metadata: if True, halt as soon as the need for metadata
        or a job code or sequence number becomes necessary
        """
        super().__init__(pref_list, problems, no_metadata)
        self.L1_date_check = VIDEO_DATE  # used in _get_date_component()

    def _get_metadata_component(self):
        """
        Returns portion of video / subfolder name based on the metadata

        Note: date time metadata found in _getDateComponent()
        """
        return get_video_metadata_component(self)


def truncate_before_unwanted_subfolder_component(pref_list: list[str]) -> list[str]:
    r"""
    truncate the preferences list to remove any subfolder element that
    contains a metadata or a job code or sequence number

    :param pref_list: subfolder prefs list
    :return: truncated list

    >>> pref_list = PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV[0]
    >>> print(truncate_before_unwanted_subfolder_component(pref_list))
    ... # doctest: +NORMALIZE_WHITESPACE
    ['Date time', 'Image date', 'YYYY', '/', '', '', 'Date time', 'Image date',
    'YYYYMMDD']
    >>> pref_list = PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV[1]
    >>> print(truncate_before_unwanted_subfolder_component(pref_list))
    ... # doctest: +NORMALIZE_WHITESPACE
    ['Date time', 'Image date', 'YYYY', '/', '', '', 'Date time', 'Image date',
    'YYYY-MM-DD']
    >>> pref_list = PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV[2]
    >>> print(truncate_before_unwanted_subfolder_component(pref_list))
    ... # doctest: +NORMALIZE_WHITESPACE
    ['Date time', 'Image date', 'YYYY', '/', '', '', 'Date time', 'Image date',
    'YYYY_MM_DD']
    >>> pref_list = PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV[3]
    >>> print(truncate_before_unwanted_subfolder_component(pref_list))
    ['Date time', 'Image date', 'YYYY']
    >>> pref_list = PHOTO_SUBFOLDER_MENU_DEFAULTS_CONV[4])
    >>> print(truncate_before_unwanted_subfolder_component(pref_list)
    ... # doctest: +NORMALIZE_WHITESPACE
    ['Date time', 'Image date', 'YYYY', '/', '', '', 'Date time', 'Image date', 'YYYY',
    'Date time', 'Image date', 'MM']
    >>> print(truncate_before_unwanted_subfolder_component([JOB_CODE, '', '',]))
    []
    >>> pl = [DATE_TIME, IMAGE_DATE, LIST_DATE_TIME_L2[11]]
    >>> print(truncate_before_unwanted_subfolder_component(pl))
    ['Date time', 'Image date', 'YYYY']
    """

    rl = [pref_list[i] for i in range(0, len(pref_list), 3)]
    truncate = -1
    for idx, value in enumerate(rl):
        if value in (METADATA, SEQUENCES, JOB_CODE):
            break
        if idx == len(rl) - 1:
            truncate = idx + 1
        elif value == SEPARATOR:
            truncate = idx

    if truncate >= 0:
        return pref_list[: truncate * 3]
    return []


def get_video_metadata_component(video: VideoSubfolder | VideoName):
    """
    Returns portion of video / subfolder name based on the metadata

    This is outside of a class definition because of the inheritance
    hierarchy.
    """

    if video.L1 == CODEC:
        v = video.rpd_file.metadata.codec()
    elif video.L1 == WIDTH:
        v = video.rpd_file.metadata.width()
    elif video.L1 == HEIGHT:
        v = video.rpd_file.metadata.height()
    elif video.L1 == FPS:
        v = video.rpd_file.metadata.frames_per_second()
    elif video.L1 == LENGTH:
        v = video.rpd_file.metadata.length()
    else:
        raise TypeError("Invalid metadata option specified")
    if video.L1 in [CODEC]:
        if video.L2 == UPPERCASE:
            v = v.upper()
        elif video.L2 == LOWERCASE:
            v = v.lower()
    if not v:
        video.problem.missing_metadata.append(_(video.L1))
    return v


class Sequences:
    """
    Stores sequence numbers and letters used in generating file names.
    """

    def __init__(
        self, downloads_today_tracker: DownloadsTodayTracker, stored_sequence_no: int
    ) -> None:
        self._session_sequence_no = 0
        self._sequence_letter = -1
        self.downloads_today_tracker = downloads_today_tracker
        self._stored_sequence_no = stored_sequence_no
        self.matched_sequences = None
        self.use_matched_sequences = False

    @property
    def session_sequence_no(self) -> int:
        if self.use_matched_sequences:
            return self.matched_sequences.session_sequence_no
        else:
            return self._session_sequence_no + 1

    @property
    def sequence_letter(self) -> int:
        if self.use_matched_sequences:
            return self.matched_sequences.sequence_letter
        else:
            return self._sequence_letter + 1

    def increment(self, uses_session_sequence_no, uses_sequence_letter) -> None:
        if uses_session_sequence_no:
            self._session_sequence_no += 1
        if uses_sequence_letter:
            self._sequence_letter += 1

    @property
    def downloads_today(self) -> int:
        if self.use_matched_sequences:
            return self.matched_sequences.downloads_today
        else:
            return self._get_downloads_today()

    def _get_downloads_today(self) -> int:
        v = self.downloads_today_tracker.get_downloads_today()
        if v == -1:
            return 1
        else:
            return v + 1

    @property
    def stored_sequence_no(self) -> int:
        if self.use_matched_sequences:
            return self.matched_sequences.stored_sequence_no
        else:
            return self._stored_sequence_no + 1

    @stored_sequence_no.setter
    def stored_sequence_no(self, value: int) -> None:
        self._stored_sequence_no = value

    def create_matched_sequences(self) -> MatchedSequences:
        return MatchedSequences(
            session_sequence_no=self._session_sequence_no + 1,
            sequence_letter=self._sequence_letter + 1,
            downloads_today=self._get_downloads_today(),
            stored_sequence_no=self._stored_sequence_no + 1,
        )



# Copyright (C) 2017-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""
Display backup preferences
"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2017-2024, Damon Lynch"

import logging
import os
from collections import defaultdict
from typing import NamedTuple

from PyQt5.QtCore import (
    QAbstractListModel,
    QModelIndex,
    QSize,
    QStorageInfo,
    Qt,
    pyqtSignal,
    pyqtSlot,
)
from PyQt5.QtGui import QIcon, QPainter, QPalette
from PyQt5.QtWidgets import (
    QCheckBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QSplitter,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QVBoxLayout,
    QWidget,
)

from raphodo.constants import (
    BackupLocationType,
    FileType,
    Roles,
    StandardFileLocations,
    ViewRowType,
)
from raphodo.devices import (
    BackupDeviceCollection,  # noqa: F401
    BackupVolumeDetails,
    DownloadingTo,
)
from raphodo.foldercombo import FolderCombo
from raphodo.prefs.preferences import Preferences
from raphodo.rpdfile import FileTypeCounter
from raphodo.storage.storage import ValidMounts, get_media_dir, get_mount_size
from raphodo.ui.destinationdisplay import adjusted_download_size, make_body_details
from raphodo.ui.devicedisplay import DeviceDisplay, DeviceView, icon_size
from raphodo.ui.panelview import QPanelView
from raphodo.ui.viewutils import (
    FlexiFrame,
    RowTracker,
    ScrollAreaNoFrame,
)


class BackupVolumeUse(NamedTuple):
    bytes_total: int
    bytes_free: int
    backup_type: BackupLocationType
    marked: FileTypeCounter
    photos_size_to_download: int
    videos_size_to_download: int


class BackupViewRow(NamedTuple):
    mount: QStorageInfo
    display_name: str
    backup_type: BackupLocationType
    os_stat_device: int


class BackupDeviceModel(QAbstractListModel):
    """
    Stores 'devices' used for backing up photos and videos.

    Want to display:
    (1) destination on local files systems
    (2) external devices, e.g., external hard drives

    Need to account for when download destination is same file system
    as backup destination.
    """

    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.raidApp = parent.rapidApp
        self.prefs = parent.prefs
        size = icon_size()
        self.removableIcon = QIcon(":icons/drive-removable-media.svg").pixmap(size)
        self.folderIcon = QIcon(":/icons/folder.svg").pixmap(size)
        self._initValues()

    def _initValues(self):
        self.rows = RowTracker()  # type: RowTracker
        self.row_id_counter = 0  # type: int
        # {row_id}
        self.headers = set()  # type: set[int]
        # path: BackupViewRow
        self.backup_devices = dict()  # type: dict[str, BackupViewRow]
        self.path_to_row_ids = defaultdict(list)  # type: dict[str, list[int]]
        self.row_id_to_path = dict()  # type: dict[int, str]

        self.marked = FileTypeCounter()
        self.photos_size_to_download = self.videos_size_to_download = 0

        # os_stat_device: set[FileType]
        self._downloading_to = defaultdict(set)  # type: DownloadingTo

    @property
    def downloading_to(self) -> DownloadingTo:
        return self._downloading_to

    @downloading_to.setter
    def downloading_to(self, downloading_to: DownloadingTo) -> None:
        self._downloading_to = downloading_to
        self.downloadSizeChanged()

    def reset(self) -> None:
        self.beginResetModel()
        self._initValues()
        self.endResetModel()

    def columnCount(self, parent=QModelIndex()):
        return 1

    def rowCount(self, parent=QModelIndex()):
        return max(len(self.rows), 1)

    def insertRows(self, position, rows=2, index=QModelIndex()):
        self.beginInsertRows(QModelIndex(), position, position + rows - 1)
        self.endInsertRows()
        return True

    def removeRows(self, position, rows=2, index=QModelIndex()):
        self.beginRemoveRows(QModelIndex(), position, position + rows - 1)
        self.endRemoveRows()
        return True

    def addBackupVolume(self, mount_details: BackupVolumeDetails) -> None:
        mount = mount_details.mount
        display_name = mount_details.name
        path = mount_details.path
        backup_type = mount_details.backup_type
        os_stat_device = mount_details.os_stat_device

        assert mount is not None
        assert display_name
        assert path
        assert backup_type

        # two rows per device: header row, and detail row
        row = len(self.rows)
        self.insertRows(position=row)
        logging.debug(
            "Adding %s to backup device display with root path %s at rows %s - %s",
            display_name,
            mount.rootPath(),
            row,
            row + 1,
        )

        for row_id in range(self.row_id_counter, self.row_id_counter + 2):
            self.row_id_to_path[row_id] = path
            self.rows[row] = row_id
            row += 1
            self.path_to_row_ids[path].append(row_id)

        header_row_id = self.row_id_counter
        self.headers.add(header_row_id)

        self.row_id_counter += 2

        self.backup_devices[path] = BackupViewRow(
            mount=mount,
            display_name=display_name,
            backup_type=backup_type,
            os_stat_device=os_stat_device,
        )

    def removeBackupVolume(self, path: str) -> None:
        """
        :param path: the value of the volume (mount's path), NOT a
        manually specified path!
        """

        row_ids = self.path_to_row_ids[path]
        header_row_id = row_ids[0]
        row = self.rows.row(header_row_id)
        logging.debug("Removing 2 rows from backup view, starting at row %s", row)
        self.rows.remove_rows(row, 2)
        self.headers.remove(header_row_id)
        del self.path_to_row_ids[path]
        del self.backup_devices[path]
        for row_id in row_ids:
            del self.row_id_to_path[row_id]
        self.removeRows(row, 2)

    def setDownloadAttributes(
        self, marked: FileTypeCounter, photos_size: int, videos_size: int, merge: bool
    ) -> None:
        """
        Set the attributes used to generate the visual display of the
        files marked to be downloaded

        :param marked: number and type of files marked for download
        :param photos_size: size in bytes of photos marked for download
        :param videos_size: size in bytes of videos marked for download
        :param merge: whether to replace or add to the current values
        """

        if not merge:
            self.marked = marked
            self.photos_size_to_download = photos_size
            self.videos_size_to_download = videos_size
        else:
            self.marked.update(marked)
            self.photos_size_to_download += photos_size
            self.videos_size_to_download += videos_size
        self.downloadSizeChanged()

    def downloadSizeChanged(self) -> None:
        # TODO possibly optimize for photo vs video rows
        for row in range(1, len(self.rows), 2):
            self.dataChanged.emit(self.index(row, 0), self.index(row, 0))

    def _download_size_by_backup_type(
        self, backup_type: BackupLocationType
    ) -> tuple[int, int]:
        """
        Include photos or videos in download size only if those file types
        are being backed up to this backup device
        :param backup_type: which file types are being backed up to this device
        :return: photos_size_to_download, videos_size_to_download
        """

        photos_size_to_download = videos_size_to_download = 0
        if backup_type != BackupLocationType.videos:
            photos_size_to_download = self.photos_size_to_download
        if backup_type != BackupLocationType.photos:
            videos_size_to_download = self.videos_size_to_download
        return photos_size_to_download, videos_size_to_download

    def data(self, index: QModelIndex, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        row = index.row()

        # check for special case where no backup devices are active
        if len(self.rows) == 0:
            if role == Qt.DisplayRole:
                return ViewRowType.header
            elif role == Roles.device_details:
                if not self.prefs.backup_files:
                    return _("Backups are not configured"), self.removableIcon
                elif self.prefs.backup_device_autodetection:
                    return _("No backup devices detected"), self.removableIcon
                else:
                    return (
                        _("Valid backup locations not yet specified"),
                        self.folderIcon,
                    )

        # at least one device  / location is being used
        if row >= len(self.rows) or row < 0:
            return None
        if row not in self.rows:
            return None

        row_id = self.rows[row]
        path = self.row_id_to_path[row_id]

        if role == Qt.DisplayRole:
            if row_id in self.headers:
                return ViewRowType.header
            else:
                return ViewRowType.content
        else:
            device = self.backup_devices[path]
            mount = device.mount

            if role == Qt.ToolTipRole:
                return path
            elif role == Roles.device_details:
                if self.prefs.backup_device_autodetection:
                    icon = self.removableIcon
                else:
                    icon = self.folderIcon
                return device.display_name, icon
            elif role == Roles.storage:
                (
                    photos_size_to_download,
                    videos_size_to_download,
                ) = self._download_size_by_backup_type(backup_type=device.backup_type)

                (
                    photos_size_to_download,
                    videos_size_to_download,
                ) = adjusted_download_size(
                    photos_size_to_download=photos_size_to_download,
                    videos_size_to_download=videos_size_to_download,
                    os_stat_device=device.os_stat_device,
                    downloading_to=self._downloading_to,
                )

                bytes_total, bytes_free = get_mount_size(mount=mount)

                return BackupVolumeUse(
                    bytes_total=bytes_total,
                    bytes_free=bytes_free,
                    backup_type=device.backup_type,
                    marked=self.marked,
                    photos_size_to_download=photos_size_to_download,
                    videos_size_to_download=videos_size_to_download,
                )

        return None

    def sufficientSpaceAvailable(self) -> bool:
        """
        Detect if each backup device has sufficient space for backing up, taking
        into accoutn situations where downloads and backups are going to the same
        partition.

        :return: False if any backup device has insufficient space, else True.
         True if there are no backup devices.
        """

        for device in self.backup_devices.values():
            (
                photos_size_to_download,
                videos_size_to_download,
            ) = self._download_size_by_backup_type(backup_type=device.backup_type)
            photos_size_to_download, videos_size_to_download = adjusted_download_size(
                photos_size_to_download=photos_size_to_download,
                videos_size_to_download=videos_size_to_download,
                os_stat_device=device.os_stat_device,
                downloading_to=self._downloading_to,
            )

            bytes_total, bytes_free = get_mount_size(mount=device.mount)
            if photos_size_to_download + videos_size_to_download >= bytes_free:
                return False
        return True


class BackupDeviceView(DeviceView):
    def __init__(self, rapidApp, parent=None) -> None:
        super().__init__(rapidApp, parent)
        self.setMouseTracking(False)
        self.entered.disconnect()
        # TODO sizeHint() ?


class BackupDeviceDelegate(QStyledItemDelegate):
    widthChanged = pyqtSignal(int)

    def __init__(self, rapidApp, parent=None) -> None:
        super().__init__(parent)
        self.rapidApp = rapidApp
        self.deviceDisplay = DeviceDisplay(parent=self)
        self.deviceDisplay.widthChanged.connect(self.widthChanged)

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        painter.save()

        x = option.rect.x()
        y = option.rect.y()
        width = option.rect.width()

        view_type = index.data(Qt.DisplayRole)  # type: ViewRowType
        if view_type == ViewRowType.header:
            display_name, icon = index.data(Roles.device_details)

            self.deviceDisplay.paint_header(
                painter=painter,
                x=x,
                y=y,
                width=width,
                icon=icon,
                display_name=display_name,
            )
        else:
            assert view_type == ViewRowType.content

            data = index.data(Roles.storage)  # type: BackupVolumeUse
            details = make_body_details(
                bytes_total=data.bytes_total,
                bytes_free=data.bytes_free,
                files_to_display=data.backup_type,
                marked=data.marked,
                photos_size_to_download=data.photos_size_to_download,
                videos_size_to_download=data.videos_size_to_download,
            )

            self.deviceDisplay.paint_body(
                painter=painter, x=x, y=y, width=width, details=details
            )

        painter.restore()

    def sizeHint(self, option: QStyleOptionViewItem, index: QModelIndex) -> QSize:
        view_type = index.data(Qt.DisplayRole)  # type: ViewRowType
        if view_type == ViewRowType.header:
            height = self.deviceDisplay.dc.device_name_height
        else:
            storage_space = index.data(Roles.storage)

            if storage_space is None:
                height = self.deviceDisplay.dc.base_height
            else:
                height = self.deviceDisplay.dc.storage_height
        return QSize(self.deviceDisplay.width(), height)


class BackupOptionsWidget(FlexiFrame):
    """
    Display backup options, such as automatic backup detection
    """

    def __init__(self, prefs: Preferences, parent, rapidApp: "RapidWindow") -> None:  # noqa: F821
        super().__init__(parent=parent)

        self.rapidApp = rapidApp
        self.prefs = prefs
        self.media_dir = get_media_dir()

        self.setBackgroundRole(QPalette.Base)
        self.setAutoFillBackground(True)

        backupLayout = QGridLayout()
        layout = self.layout()
        layout.addLayout(backupLayout)
        self.setLayout(layout)

        self.backupExplanation = QLabel(
            _(
                "You can have your photos and videos backed up to "
                "multiple locations as they are downloaded, e.g. "
                "external hard drives."
            )
        )
        self.backupExplanation.setWordWrap(True)
        self.backupExplanation.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Minimum)

        self.backup = QCheckBox(_("Back up photos and videos when downloading"))
        self.backup.setChecked(self.prefs.backup_files)
        self.backup.stateChanged.connect(self.backupChanged)

        checkbox_width = self.backup.style().pixelMetric(QStyle.PM_IndicatorWidth)

        self.autoBackup = QCheckBox(_("Automatically detect backup devices"))
        self.autoBackup.setChecked(self.prefs.backup_device_autodetection)
        self.autoBackup.stateChanged.connect(self.autoBackupChanged)

        self.folderExplanation = QLabel(
            _(
                "Specify the folder in which backups are stored on the "
                "device.<br><br>"
                "<i>Note: the presence of a folder with this name "
                "is used to determine if the device is used for backups. "
                "For each device you wish to use for backing up to, "
                "create a folder in it with one of these folder names. "
                "By adding both folders, the same device can be used "
                "to back up both photos and videos.</i>"
            )
        )
        self.folderExplanation.setWordWrap(True)
        self.folderExplanation.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Minimum)
        # Unless this next call is made, for some reason the widget is too high! :-(
        self.folderExplanation.setContentsMargins(0, 0, 1, 0)

        self.photoFolderNameLabel = QLabel(_("Photo folder name:"))
        self.photoFolderName = QLineEdit()
        self.photoFolderName.setText(self.prefs.photo_backup_identifier)
        self.photoFolderName.editingFinished.connect(self.photoFolderIdentifierChanged)

        self.videoFolderNameLabel = QLabel(_("Video folder name:"))
        self.videoFolderName = QLineEdit()
        self.videoFolderName.setText(self.prefs.video_backup_identifier)
        self.videoFolderName.editingFinished.connect(self.videoFolderIdentifierChanged)

        self.autoBackupExampleBox = QGroupBox(_("Example:"))
        self.autoBackupExample = QLabel()

        autoBackupExampleBoxLayout = QHBoxLayout()
        autoBackupExampleBoxLayout.addWidget(self.autoBackupExample)

        self.autoBackupExampleBox.setLayout(autoBackupExampleBoxLayout)

        valid_mounts = ValidMounts(only_external_mounts=True)

        self.manualLocationExplanation = QLabel(
            _("If you disable automatic detection, choose the exact backup locations.")
        )
        self.manualLocationExplanation.setWordWrap(True)
        self.manualLocationExplanation.setSizePolicy(
            QSizePolicy.Ignored, QSizePolicy.Minimum
        )
        # Translators: the word 'location' is optional in your translation. The left
        # side of the folder chooser combo box will always line up with the left side of
        # the the text entry boxes where the user can enter the photo folder name and
        # the video folder name. See
        # http://damonlynch.net/rapid/documentation/thumbnails/backup.png
        self.photoLocationLabel = QLabel(_("Photo backup location:"))
        self.photoLocationLabel.setWordWrap(True)
        if self.rapidApp.is_wsl2 and not self.rapidApp.wsl_drives_probed:
            # flag that once Windows drives volume names are known, update these
            # combo box values
            self.rapidApp.wsl_backup_drives_refresh_needed = True
        self.photoLocation = FolderCombo(
            self,
            prefs=self.prefs,
            file_type=FileType.photo,
            file_chooser_title=_("Select Photo Backup Location"),
            special_dirs=(StandardFileLocations.pictures,),
            valid_mounts=valid_mounts,
        )
        self.photoLocation.setPath(self.prefs.backup_photo_location)
        self.photoLocation.pathChosen.connect(self.photoPathChosen)

        # Translators: the word 'location' is optional in your translation. The left
        # side of the folder chooser combo box will always line up with the left side of
        # the the text entry boxes where the user can enter the photo folder name and
        # the video folder name. See
        # http://damonlynch.net/rapid/documentation/thumbnails/backup.png
        self.videoLocationLabel = QLabel(_("Video backup location:"))
        self.videoLocationLabel.setWordWrap(True)
        self.videoLocation = FolderCombo(
            self,
            prefs=self.prefs,
            file_type=FileType.video,
            file_chooser_title=_("Select Video Backup Location"),
            special_dirs=(StandardFileLocations.videos,),
            valid_mounts=valid_mounts,
        )
        self.videoLocation.setPath(self.prefs.backup_video_location)
        self.videoLocation.pathChosen.connect(self.videoPathChosen)

        backupLayout.addWidget(self.backupExplanation, 0, 0, 1, 4)
        backupLayout.addWidget(self.backup, 1, 0, 1, 4)
        backupLayout.addWidget(self.autoBackup, 2, 1, 1, 3)
        backupLayout.addWidget(self.folderExplanation, 3, 2, 1, 2)
        backupLayout.addWidget(self.photoFolderNameLabel, 4, 2, 1, 1)
        backupLayout.addWidget(self.photoFolderName, 4, 3, 1, 1)
        backupLayout.addWidget(self.videoFolderNameLabel, 5, 2, 1, 1)
        backupLayout.addWidget(self.videoFolderName, 5, 3, 1, 1)
        backupLayout.addWidget(self.autoBackupExampleBox, 6, 2, 1, 2)
        backupLayout.addWidget(
            self.manualLocationExplanation, 7, 1, 1, 3, Qt.AlignBottom
        )
        backupLayout.addWidget(self.photoLocationLabel, 8, 1, 1, 2)
        backupLayout.addWidget(self.photoLocation, 8, 3, 1, 1)
        backupLayout.addWidget(self.videoLocationLabel, 9, 1, 1, 2)
        backupLayout.addWidget(self.videoLocation, 9, 3, 1, 1)

        backupLayout.setColumnMinimumWidth(0, checkbox_width)
        backupLayout.setColumnMinimumWidth(1, checkbox_width)

        backupLayout.setRowMinimumHeight(7, checkbox_width * 2)

        layout.addStretch()

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.setBackupButtonHighlight()

        # Group controls to enable / disable sets of them
        self._backup_controls_type = (self.autoBackup,)
        self._backup_controls_auto = (
            self.folderExplanation,
            self.photoFolderNameLabel,
            self.photoFolderName,
            self.videoFolderNameLabel,
            self.videoFolderName,
            self.autoBackupExampleBox,
        )
        self._backup_controls_manual = (
            self.manualLocationExplanation,
            self.photoLocationLabel,
            self.photoLocation,
            self.videoLocationLabel,
            self.videoLocation,
        )
        self.updateExample()
        self.enableControlsByBackupType()

    @pyqtSlot(int)
    def backupChanged(self, state: int) -> None:
        backup = state == Qt.Checked
        logging.info("Setting backup while downloading to %s", backup)
        self.prefs.backup_files = backup
        self.setBackupButtonHighlight()
        self.enableControlsByBackupType()
        self.rapidApp.resetupBackupDevices()

    @pyqtSlot(int)
    def autoBackupChanged(self, state: int) -> None:
        autoBackup = state == Qt.Checked
        logging.info("Setting automatically detect backup devices to %s", autoBackup)
        self.prefs.backup_device_autodetection = autoBackup
        self.setBackupButtonHighlight()
        self.enableControlsByBackupType()
        self.rapidApp.resetupBackupDevices()

    @pyqtSlot(str)
    def photoPathChosen(self, path: str) -> None:
        logging.info("Setting backup photo location to %s", path)
        self.prefs.backup_photo_location = path
        self.setBackupButtonHighlight()
        self.rapidApp.resetupBackupDevices()

    @pyqtSlot(str)
    def videoPathChosen(self, path: str) -> None:
        logging.info("Setting backup video location to %s", path)
        self.prefs.backup_video_location = path
        self.setBackupButtonHighlight()
        self.rapidApp.resetupBackupDevices()

    @pyqtSlot()
    def photoFolderIdentifierChanged(self) -> None:
        name = self.photoFolderName.text()
        logging.info("Setting backup photo folder name to %s", name)
        self.prefs.photo_backup_identifier = name
        self.setBackupButtonHighlight()
        self.rapidApp.resetupBackupDevices()

    @pyqtSlot()
    def videoFolderIdentifierChanged(self) -> None:
        name = self.videoFolderName.text()
        logging.info("Setting backup video folder name to %s", name)
        self.prefs.video_backup_identifier = name
        self.setBackupButtonHighlight()
        self.rapidApp.resetupBackupDevices()

    def updateExample(self) -> None:
        """
        Update the example paths in the backup panel
        """

        if (
            self.autoBackup.isChecked()
            and hasattr(self.rapidApp, "backup_devices")
            and len(self.rapidApp.backup_devices)
        ):
            drives = self.rapidApp.backup_devices.sample_device_paths()
        else:
            # Translators: this value is used as an example device when automatic backup
            # device detection is enabled. You should translate this.
            drive1 = os.path.join(self.media_dir, _("drive1"))
            # Translators: this value is used as an example device when automatic backup
            # device detection is enabled. You should translate this.
            drive2 = os.path.join(self.media_dir, _("drive2"))
            drives = (
                os.path.join(path, identifier)
                for path, identifier in (
                    (drive1, self.prefs.photo_backup_identifier),
                    (drive2, self.prefs.photo_backup_identifier),
                    (drive2, self.prefs.video_backup_identifier),
                )
            )
        paths = "\n".join(drives)
        self.autoBackupExample.setText(paths)

    def setBackupButtonHighlight(self) -> None:
        """
        Indicate error status in GUI by highlighting Backup button.

        Do so only if doing manual backups and there is a problem with one of the paths
        """

        self.rapidApp.backupButton.setHighlighted(
            self.prefs.backup_files
            and not self.prefs.backup_device_autodetection
            and (self.photoLocation.invalid_path or self.videoLocation.invalid_path)
        )

    def enableControlsByBackupType(self) -> None:
        """
        Enable or disable backup controls depending on what the user
        has enabled.
        """

        backupsEnabled = self.backup.isChecked()
        autoEnabled = backupsEnabled and self.autoBackup.isChecked()
        manualEnabled = not autoEnabled and backupsEnabled

        for widget in self._backup_controls_type:
            widget.setEnabled(backupsEnabled)
        for widget in self._backup_controls_manual:
            widget.setEnabled(manualEnabled)
        for widget in self._backup_controls_auto:
            widget.setEnabled(autoEnabled)

    def updateLocationCombos(self) -> None:
        """
        Update backup location comboboxes in case directory status has changed.
        """
        for combo in self.photoLocation, self.videoLocation:
            combo.refreshFolderList()


class BackupPanel(ScrollAreaNoFrame):
    """
    Backup preferences widget, for photos and video backups while
    downloading.
    """

    def __init__(self, parent) -> None:
        super().__init__(name="backupPanel", parent=parent)

        assert parent is not None
        self.rapidApp = parent
        self.prefs = self.rapidApp.prefs  # type: Preferences
        self.setObjectName("backupPanelScrollArea")

        self.backupDevices = BackupDeviceModel(parent=self)

        self.backupStoragePanel = QPanelView(
            label=_("Projected Backup Storage Use"),
        )
        self.backupStoragePanel.setObjectName("backupStoragePanel")

        self.backupOptionsPanel = QPanelView(
            label=_("Backup Options"),
        )
        self.backupOptionsPanel.setObjectName("backupOptionsPanel")

        self.backupDevicesView = BackupDeviceView(rapidApp=self.rapidApp, parent=self)
        self.backupDevicesView.setObjectName("backupDevicesView")
        self.backupDevicesView.setModel(self.backupDevices)
        self.backupDevicesView.setItemDelegate(
            BackupDeviceDelegate(rapidApp=self.rapidApp)
        )
        self.backupDevicesView.setSizePolicy(
            QSizePolicy.MinimumExpanding, QSizePolicy.Fixed
        )
        self.backupOptionsPanel.setSizePolicy(
            QSizePolicy.Preferred, QSizePolicy.MinimumExpanding
        )

        self.backupOptions = BackupOptionsWidget(
            prefs=self.prefs, parent=self, rapidApp=self.rapidApp
        )

        self.backupOptionsPanel.addWidget(self.backupOptions)
        self.backupStoragePanel.addWidget(self.backupDevicesView)

        self.verticalScrollBarVisible.connect(
            self.backupDevicesView.containerVerticalScrollBar
        )
        self.verticalScrollBarVisible.connect(
            self.backupOptions.containerVerticalScrollBar
        )
        self.horizontalScrollBarVisible.connect(
            self.backupOptions.containerHorizontalScrollBar
        )
        widget = QWidget()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(QSplitter().handleWidth())
        widget.setLayout(layout)
        layout.addWidget(self.backupStoragePanel)
        layout.addWidget(self.backupOptionsPanel)
        self.setWidget(widget)
        self.setWidgetResizable(True)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)

    def updateExample(self) -> None:
        """
        Update the example paths in the backup panel
        """

        self.backupOptions.updateExample()

    def updateLocationCombos(self) -> None:
        """
        Update backup location comboboxes in case directory status has changed.
        """

        self.backupOptions.updateLocationCombos()

    def addBackupVolume(self, mount_details: BackupVolumeDetails) -> None:
        self.backupDevices.addBackupVolume(mount_details=mount_details)
        self.backupDevicesView.updateGeometry()

    def removeBackupVolume(self, path: str) -> None:
        self.backupDevices.removeBackupVolume(path=path)
        self.backupDevicesView.updateGeometry()

    def resetBackupDisplay(self) -> None:
        self.backupDevices.reset()
        self.backupDevicesView.updateGeometry()

    def setupBackupDisplay(self) -> None:
        """
        Sets up the backup view list regardless of whether backups
        are manual specified by the user, or auto-detection is on
        """

        if not self.prefs.backup_files:
            logging.debug("No backups configured: no backup destinations to display")
            return

        backup_devices = self.rapidApp.backup_devices  # type: BackupDeviceCollection
        if self.prefs.backup_device_autodetection:
            for path in backup_devices:
                self.backupDevices.addBackupVolume(
                    mount_details=backup_devices.get_backup_volume_details(path=path)
                )
        else:
            # manually specified backup paths
            try:
                mounts = backup_devices.get_manual_mounts()
                if mounts is None:
                    return

                self.backupDevices.addBackupVolume(mount_details=mounts[0])
                if len(mounts) > 1:
                    self.backupDevices.addBackupVolume(mount_details=mounts[1])
            except Exception:
                logging.exception(
                    "An unexpected error occurred when adding backup destinations. "
                    "Exception:"
                )
        self.backupDevicesView.updateGeometry()

    def setDownloadAttributes(
        self, marked: FileTypeCounter, photos_size: int, videos_size: int, merge: bool
    ) -> None:
        """
        Set the attributes used to generate the visual display of the
        files marked to be downloaded

        :param marked: number and type of files marked for download
        :param photos_size: size in bytes of photos marked for download
        :param videos_size: size in bytes of videos marked for download
        :param merge: whether to replace or add to the current values
        """

        self.backupDevices.setDownloadAttributes(
            marked=marked, photos_size=photos_size, videos_size=videos_size, merge=merge
        )

    def sufficientSpaceAvailable(self) -> bool:
        """
        Check to see that there is sufficient space with which to perform a download.

        :return: True or False value if sufficient space. Will always return True if
         backups are disabled or there are no backup devices.
        """
        if self.prefs.backup_files:
            return self.backupDevices.sufficientSpaceAvailable()
        else:
            return True

    def setDownloadingTo(self, downloading_to: DownloadingTo) -> None:
        self.backupDevices.downloading_to = downloading_to



#!/usr/bin/env python3

# Copyright (C) 2015-2024 Damon Lynch <damonlynch@gmail.com>

# This file is part of Rapid Photo Downloader.
#
# Rapid Photo Downloader is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rapid Photo Downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rapid Photo Downloader.  If not,
# see <http://www.gnu.org/licenses/>.

"""Starts load balancer used for thumbnail extraction and caching"""

__author__ = "Damon Lynch"
__copyright__ = "Copyright 2015-2024, Damon Lynch"

from raphodo.interprocess import LoadBalancer, LoadBalancerWorkerManager


class ThumbnailLoadBalancerWorkerManager(LoadBalancerWorkerManager):
    """
    Manages thumbnail extractors
    """

    def __init__(
        self, no_workers: int, backend_port: int, sink_port: int, logging_port: int
    ) -> None:
        super().__init__(no_workers, backend_port, sink_port, logging_port)
        self._process_name = "Thumbnail Load Balancer Manager"
        self._process_to_run = "thumbnailextractor.py"


class ThumbnailLoadBalancer(LoadBalancer):
    """
    Manages the thumbnail load balancer
    """

    def __init__(self) -> None:
        super().__init__("Thumbnail", ThumbnailLoadBalancerWorkerManager)


if __name__ == "__main__":
    loadbalancer = ThumbnailLoadBalancer()
